# -*- coding: utf-8 -*-
'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Joy Cheung <joycheung1994@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/master/LICENSE.txt
for license details.
'''

import thermosteam as tmo
from thermosteam.utils import chemicals_user
# import os
# os.chdir("C:/Users/joy_c/Dropbox/PhD/Research/QSD/codes_developing/QSDsan")
from qsdsan import Components, Process, Processes
from ..utils import data_path

__all__ = ('cmps_asm2d', 'asm2d', 'ASM2d')

data_path += 'process_data/_asm2d.tsv'
# import os
# data_path = "qsdsan/data/process_data/_asm2d.tsv"

############# Components with default notation #############
cmps = Components.load_default()

S_A = cmps.S_Ac.copy('S_A')
S_ALK = cmps.S_CO3.copy('S_ALK')      # measured as g C
S_F = cmps.S_F.copy('S_F')
S_I = cmps.S_U_E.copy('S_I')
S_N2 = cmps.S_N2.copy('S_N2')
S_NH4 = cmps.S_NH4.copy('S_NH4')
S_NO3 = cmps.S_NO3.copy('S_NO3')
S_O2 = cmps.S_O2.copy('S_O2')
S_PO4 = cmps.S_PO4.copy('S_PO4')

X_AUT = cmps.X_AOO.copy('X_AUT')
X_H = cmps.X_OHO.copy('X_H')
X_I = cmps.X_U_OHO_E.copy('X_I')
X_MeOH = cmps.X_FeOH.copy('X_MeOH')
X_MeP = cmps.X_FePO4.copy('X_MeP')
X_PAO = cmps.X_PAO.copy('X_PAO')
X_PHA = cmps.X_PAO_PHA.copy('X_PHA')
X_PP = cmps.X_PAO_PP_Lo.copy('X_PP')
X_S = cmps.X_B_Subst.copy('X_S')

S_I.i_N = 0.01
S_F.i_N = 0.03
X_I.i_N = 0.02
X_S.i_N = 0.04
X_H.i_N = X_PAO.i_N = X_AUT.i_N = 0.07

S_I.i_P = 0.00
S_F.i_P = 0.01
X_I.i_P = 0.01
X_S.i_P = 0.01
X_H.i_P = X_PAO.i_P = X_AUT.i_P = 0.02

X_I.i_mass = 0.75
X_S.i_mass = 0.75
X_H.i_mass = X_PAO.i_mass = X_AUT.i_mass = 0.9

cmps_asm2d = Components([S_O2, S_F, S_A, S_NH4, S_NO3, S_PO4, S_I, S_ALK, S_N2,
                         X_I, X_S, X_H, X_PAO, X_PP, X_PHA, X_AUT, X_MeOH, X_MeP,
                         cmps.H2O])

cmps_asm2d.compile()
tmo.settings.set_thermo(cmps_asm2d)

############ Processes in ASM2d #################
params = ('f_SI', 'Y_H', 'f_XI_H', 'Y_PAO', 'Y_PO4', 'Y_PHA', 'f_XI_PAO', 'Y_A', 'f_XI_AUT',
          'K_h', 'eta_NO3', 'eta_fe', 'K_O2', 'K_NO3', 'K_X',
          'mu_H', 'q_fe', 'eta_NO3_H', 'b_H', 'K_O2_H', 'K_F', 'K_fe', 'K_A_H',
          'K_NO3_H', 'K_NH4_H', 'K_P_H', 'K_ALK_H',
          'q_PHA', 'q_PP', 'mu_PAO', 'eta_NO3_PAO', 'b_PAO', 'b_PP', 'b_PHA',
          'K_O2_PAO', 'K_NO3_PAO', 'K_A_PAO', 'K_NH4_PAO', 'K_PS',' K_P_PAO',
          'K_ALK_PAO', 'K_PP', 'K_MAX', 'K_IPP', 'K_PHA',
          'mu_AUT', 'b_AUT', 'K_O2_AUT', 'K_NH4_AUT', 'K_ALK_AUT', 'K_P_AUT',
          'k_PRE', 'k_RED', 'K_ALK_PRE')

asm2d = Processes.load_from_file(data_path,
                                 conserved_for=('COD', 'N', 'P', 'charge'),
                                 parameters=params,
                                 compile=False)

p12 = Process('anox_storage_PP',
              'S_PO4 + [Y_PHA]X_PHA + [?]S_NO3 -> X_PP + [?]S_N2 + [?]S_NH4 + [?]S_ALK',
              ref_component='X_PP',
              rate_equation='q_PP * S_O2/(K_O2_PAO+S_O2) * S_PO4/(K_PS+S_PO4) * S_ALK/(K_ALK_PAO+S_ALK) * (X_PHA/X_PAO)/(K_PHA+X_PHA/X_PAO) * (K_MAX-X_PP/X_PAO)/(K_IPP+K_MAX-X_PP/X_PAO) * X_PAO * eta_NO3_PAO * K_O2_PAO/S_O2 * S_NO3/(K_NO3_PAO+S_NO3)',
              parameters=('Y_PHA', 'q_PP', 'K_O2_PAO', 'K_PS', 'K_ALK_PAO', 'K_PHA', 'eta_NO3_PAO', 'K_IPP', 'K_NO3_PAO'),
              conserved_for=('COD', 'N', 'P', 'NOD', 'charge'))

p14 = Process('PAO_anox_growth',
             '[1/Y_PAO]X_PHA + [?]S_NO3 + [?]S_PO4 -> X_PAO + [?]S_N2 + [?]S_NH4  + [?]S_ALK',
             ref_component='X_PAO',
             rate_equation='mu_PAO * S_O2/(K_O2_PAO + S_O2) * S_NH4/(K_NH4_PAO + S_NH4) * S_PO4/(K_P_PAO + S_PO4) * S_ALK/(K_ALK_PAO + S_ALK) * (X_PHA/X_PAO)/(K_PHA + X_PHA/X_PAO) * X_PAO * eta_NO3_PAO * K_O2_PAO/S_O2 * S_NO3/(K_NO3_PAO + S_NO3)',
             parameters=('Y_PAO', 'mu_PAO', 'K_O2_PAO', 'K_NH4_PAO', 'K_P_PAO', 'K_ALK_PAO', 'K_PHA', 'eta_NO3_PAO', 'K_NO3_PAO'),
             conserved_for=('COD', 'N', 'P', 'NOD', 'charge'))

asm2d.extend([p12, p14])
asm2d.compile()

# ASM2d typical values at 20 degree C
asm2d.set_parameters(
    f_SI = 0,                    # production of soluble inerts in hydrolysis = 0.0 gCOD/gCOD
    Y_H = 0.625,                 # heterotrophic yield = 0.625 gCOD/gCOD
    f_XI_H=0.1,                  # fraction of inert COD generated in heterotrophic biomass lysis = 0.1 gCOD/gCOD
    Y_PAO = 0.625,               # PAO yield = 0.625 gCOD/gCOD
    Y_PO4 = 0.4,                 # PP requirement (PO4 release) per PHA stored = 0.4 gP/gCOD
    Y_PHA = 0.2,                 # PHA requirement for PP storage = 0.2 gCOD/gP
    f_XI_PAO=0.1,                # fraction of inert COD generated in PAO biomass lysis = 0.1 gCOD/gCOD
    Y_A = 0.24,                  # autotrophic yield = 0.24 gCOD/gN
    f_XI_AUT=0.1,                # fraction of inert COD generated in autotrophic biomass lysis = 0.1 gCOD/gCOD
    K_h = 3,                     # hydrolysis rate constant = 3.0 d^(-1)
    eta_NO3 = 0.6,               # reduction factor for anoxic hydrolysis = 0.6
    eta_fe = 0.4,                # anaerobic hydrolysis reduction factor = 0.4
    K_O2 = 0.2,                  # O2 half saturation coefficient of hydrolysis = 0.2 mgO2/L
    K_NO3 = 0.5,                 # nitrate half saturation coefficient of hydrolysis = 0.5 mgN/L
    K_X = 0.1,                   # slowly biodegradable substrate half saturation coefficient for hydrolysis = 0.1 gCOD/gCOD
    mu_H = 6,                    # heterotrophic maximum specific growth rate = 6.0 d^(-1)
    q_fe = 3,                    # fermentation maximum rate = 3.0 d^(-1)
    eta_NO3_H = 0.8,             # denitrification reduction factor for heterotrophic growth = 0.8
    b_H = 0.4,                   # lysis and decay rate constant = 0.4 d^(-1)
    K_O2_H=0.2,                  # O2 half saturation coefficient of heterotrophs = 0.2 mgO2/L
    K_F = 4,                     # fermentable substrate half saturation coefficient for heterotrophic growth = 4.0 mgCOD/L
    K_fe = 4,                    # fermentable substrate half saturation coefficient for fermentation = 4.0 mgCOD/L
    K_A_H = 4,                   # VFA half saturation coefficient for heterotrophs = 4.0 mgCOD/L
    K_NO3_H=0.5,                 # nitrate half saturation coefficient = 0.5 mgN/L
    K_NH4_H = 0.05,              # ammonium (nutrient) half saturation coefficient for heterotrophs = 0.05 mgN/L
    K_P_H = 0.01,                # phosphorus (nutrient) half saturation coefficient for heterotrophs = 0.01 mgP/L
    K_ALK_H = 0.1*12,            # alkalinity half saturation coefficient for heterotrophs = 0.1 mol(HCO3-)/m^3 = 1.2 gC/m^3
    q_PHA = 3,                   # rate constant for storage of PHA = 3.0 d^(-1)
    q_PP = 1.5,                  # rate constant for storage of PP = 1.5 d^(-1)
    mu_PAO = 1,                  # PAO maximum specific growth rate = 1.0 d^(-1)
    eta_NO3_PAO=0.6,             # denitrification reduction factor for PAO growth = 0.8
    b_PAO = 0.2,                 # PAO lysis rate = 0.2 d^(-1)
    b_PP = 0.2,                  # PP lysis rate = 0.2 d^(-1)
    b_PHA = 0.2,                 # PHA lysis rate = 0.2 d^(-1)
    K_O2_PAO=0.2,                # O2 half saturation coefficient for PAOs = 0.2 mgO2/L
    K_NO3_PAO=0.5,               # nitrate half saturation coefficient for PAOs = 0.5 mgN/L
    K_A_PAO=4.0,                 # VFA half saturation coefficient for PAOs = 4.0 mgCOD/L
    K_NH4_PAO=0.05,              # ammonium (nutrient) half saturation coefficient for PAOs = 0.05 mgN/L
    K_PS = 0.2,                  # phosphorus half saturation coefficient for storage of PP = 0.2 mgP/L
    K_P_PAO=0.01,                # phosphorus (nutrient) half saturation coefficient for heterotrophs = 0.01 mgP/L
    K_ALK_PAO=0.1*12,            # alkalinity half saturation coefficient for PAOs = 0.1 mol(HCO3-)/m^3 = 1.2 gC/m^3
    K_PP = 0.01,                 # PP half saturation coefficient for storage of PHA = 0.01 gP/gCOD (?). gCOD/gCOD in GPS-X
    K_MAX = 0.34,                # maximum ratio of X_PP/X_PAO = 0.34 gX_PP/gX_PAO
    K_IPP = 0.02,                # inhibition coefficient for PP storage = 0.02 gP/gCOD
    K_PHA = 0.01,                # PHA half saturation coefficient = 0.01 gCOD/gCOD
    mu_AUT = 1,                  # autotrophic maximum specific growth rate = 1.0 d^(-1)
    b_AUT = 0.15,                # autotrophic decay rate = 0.15 d^(-1)
    K_O2_AUT = 0.5,              # O2 half saturation coefficient for autotrophic growth = 0.5 mgO2/L
    K_NH4_AUT = 1,               # ammonium (substrate) half saturation coefficient for autotrophic growth = 1.0 mgN/L
    K_ALK_AUT=0.5*12,            # alkalinity half saturation coefficient for autotrophic growth = 0.5 mol(HCO3-)/m^3 = 6.0 gC/m^3
    K_P_AUT=0.01,                # phosphorus (nutrient) half saturation coefficient for autotrophic growth = 0.01 mgP/L
    k_PRE = 1,                   # phosphorus precipitation with MeOH rate constant = 1.0 m^3/g/d
    k_RED = 0.6,                 # redissoluation of phosphates rate constant = 0.6 d^(-1)
    K_ALK_PRE=0.5*12             # alkalinity half saturation coefficient for phosphate precipitation = 0.5 mol(HCO3-)/m^3 = 6.0 gC/m^3
    )

# ASM2d typical values at 10 degree C
# asm2d.set_parameters(
#     f_SI = 0,                    # production of soluble inerts in hydrolysis = 0.0 gCOD/gCOD
#     Y_H = 0.625,                 # heterotrophic yield = 0.625 gCOD/gCOD
#     f_XI_H=0.1,                  # fraction of inert COD generated in heterotrophic biomass lysis = 0.1 gCOD/gCOD
#     Y_PAO = 0.625,               # PAO yield = 0.625 gCOD/gCOD
#     Y_PO4 = 0.4,                 # PP requirement (PO4 release) per PHA stored = 0.4 gP/gCOD
#     Y_PHA = 0.2,                 # PHA requirement for PP storage = 0.2 gCOD/gP
#     f_XI_PAO=0.1,                # fraction of inert COD generated in PAO biomass lysis = 0.1 gCOD/gCOD
#     Y_A = 0.24,                  # autotrophic yield = 0.24 gCOD/gN
#     f_XI_AUT=0.1,                # fraction of inert COD generated in autotrophic biomass lysis = 0.1 gCOD/gCOD
#     K_h = 2,                     # hydrolysis rate constant = 2.0 d^(-1)
#     eta_NO3 = 0.6,               # reduction factor for anoxic hydrolysis = 0.6
#     eta_fe = 0.4,                # anaerobic hydrolysis reduction factor = 0.4
#     K_O2 = 0.2,                  # O2 half saturation coefficient of hydrolysis = 0.2 mgO2/L
#     K_NO3 = 0.5,                 # nitrate half saturation coefficient of hydrolysis = 0.5 mgN/L
#     K_X = 0.1,                   # slowly biodegradable substrate half saturation coefficient for hydrolysis = 0.1 gCOD/gCOD
#     mu_H = 3,                    # heterotrophic maximum specific growth rate = 3.0 d^(-1)
#     q_fe = 1.5,                  # fermentation maximum rate = 1.5 d^(-1)
#     eta_NO3_H = 0.8,             # denitrification reduction factor for heterotrophic growth = 0.8
#     b_H = 0.4,                   # lysis and decay rate constant = 0.4 d^(-1)
#     K_O2_H=0.2,                  # O2 half saturation coefficient of heterotrophs = 0.2 mgO2/L
#     K_F = 4,                     # fermentable substrate half saturation coefficient for heterotrophic growth = 4.0 mgCOD/L
#     K_fe = 4,                    # fermentable substrate half saturation coefficient for fermentation = 4.0 mgCOD/L
#     K_A_H = 4,                   # VFA half saturation coefficient for heterotrophs = 4.0 mgCOD/L
#     K_NO3_H=0.5,                 # nitrate half saturation coefficient = 0.5 mgN/L
#     K_NH4_H = 0.05,              # ammonium (nutrient) half saturation coefficient for heterotrophs = 0.05 mgN/L
#     K_P_H = 0.01,                # phosphorus (nutrient) half saturation coefficient for heterotrophs = 0.01 mgP/L
#     K_ALK_H = 0.1*12,            # alkalinity half saturation coefficient for heterotrophs = 0.1 mol(HCO3-)/m^3 = 1.2 gC/m^3
#     q_PHA = 2,                   # rate constant for storage of PHA = 2.0 d^(-1)
#     q_PP = 1.0,                  # rate constant for storage of PP = 1.0 d^(-1)
#     mu_PAO = 0.67,               # PAO maximum specific growth rate = 0.67 d^(-1)
#     eta_NO3_PAO=0.6,             # denitrification reduction factor for PAO growth = 0.8
#     b_PAO = 0.1,                 # PAO lysis rate = 0.1 d^(-1)
#     b_PP = 0.1,                  # PP lysis rate = 0.1 d^(-1)
#     b_PHA = 0.1,                 # PHA lysis rate = 0.1 d^(-1)
#     K_O2_PAO=0.2,                # O2 half saturation coefficient for PAOs = 0.2 mgO2/L
#     K_NO3_PAO=0.5,               # nitrate half saturation coefficient for PAOs = 0.5 mgN/L
#     K_A_PAO=4.0,                 # VFA half saturation coefficient for PAOs = 4.0 mgCOD/L
#     K_NH4_PAO=0.05,              # ammonium (nutrient) half saturation coefficient for PAOs = 0.05 mgN/L
#     K_PS = 0.2,                  # phosphorus half saturation coefficient for storage of PP = 0.2 mgP/L
#     K_P_PAO=0.01,                # phosphorus (nutrient) half saturation coefficient for heterotrophs = 0.01 mgP/L
#     K_ALK_PAO=0.1*12,            # alkalinity half saturation coefficient for PAOs = 0.1 mol(HCO3-)/m^3 = 1.2 gC/m^3
#     K_PP = 0.01,                 # PP half saturation coefficient for storage of PHA = 0.01 gP/gCOD (?). gCOD/gCOD in GPS-X
#     K_MAX = 0.34,                # maximum ratio of X_PP/X_PAO = 0.34 gX_PP/gX_PAO
#     K_IPP = 0.02,                # inhibition coefficient for PP storage = 0.02 gP/gCOD
#     K_PHA = 0.01,                # PHA half saturation coefficient = 0.01 gCOD/gCOD
#     mu_AUT = 0.35,               # autotrophic maximum specific growth rate = 0.35 d^(-1)
#     b_AUT = 0.05,                # autotrophic decay rate = 0.05 d^(-1)
#     K_O2_AUT = 0.5,              # O2 half saturation coefficient for autotrophic growth = 0.5 mgO2/L
#     K_NH4_AUT = 1,               # ammonium (substrate) half saturation coefficient for autotrophic growth = 1.0 mgN/L
#     K_ALK_AUT=0.5*12,            # alkalinity half saturation coefficient for autotrophic growth = 0.5 mol(HCO3-)/m^3 = 6.0 gC/m^3
#     K_P_AUT=0.01,                # phosphorus (nutrient) half saturation coefficient for autotrophic growth = 0.01 mgP/L
#     k_PRE = 1,                   # phosphorus precipitation with MeOH rate constant = 1.0 m^3/g/d
#     k_RED = 0.6,                 # redissoluation of phosphates rate constant = 0.6 d^(-1)
#     K_ALK_PRE=0.5*12             # alkalinity half saturation coefficient for phosphate precipitation = 0.5 mol(HCO3-)/m^3 = 6.0 gC/m^3
#     )

@chemicals_user
class ASM2d(Processes):
    '''
    Activated Sludge Model No. 2d in orginal notation.

    Parameters
    ----------
    components: class:`CompiledComponents`, optional
        Components corresponding to each entry in the stoichiometry array,
        defaults to thermosteam.settings.chemicals.
    iN_SI : float, optional
        Nitrogen content of inert soluble COD, in [g N/g COD]. The default is 0.01.
    iN_SF : float, optional
        Nitrogen content of fermentable substrate, in [g N/g COD]. The default is 0.03.
    iN_XI : float, optional
        Nitrogen content of inert particulate COD, in [g N/g COD]. The default is 0.02.
    iN_XS : float, optional
        Nitrogen content of slowly biodegradable substrate, in [g N/g COD]. The default
        is 0.04.
    iN_BM : float, optional
        Nitrogen content of biomass, in [g N/g COD]. The default is 0.07.
    iP_SI : float, optional
        Phosphorus content of inert soluble COD, in [g P/g COD]. The default is 0.0.
    iP_SF : float, optional
        Phosphorus content of fermentable substrate, in [g P/g COD]. The default
        is 0.01.
    iP_XI : float, optional
        Phosphorus content of inert particulate COD, in [g P/g COD]. The default
        is 0.01.
    iP_XS : float, optional
        Phosphorus content of slowly biodegradable substrate, in [g P/g COD]. The
        default is 0.01.
    iP_BM : float, optional
        Phosphorus content of biomass, in [g P/g COD]. The default is 0.02.
    iTSS_XI : float, optional
        TSS to COD ratio for inert particulate COD, in [g TSS/g COD]. The default
        is 0.75.
    iTSS_XS : float, optional
        TSS to COD ratio for slowly biodegradable substrate, in [g TSS/g COD]. The
        default is 0.75.
    iTSS_BM : float, optional
        TSS to COD ratio for biomass, in [g TSS/g COD]. The default is 0.9.
    f_SI : float, optional
        Production of soluble inerts in hydrolysis, in [g COD/g COD]. The default
        is 0.0.
    Y_H : float, optional
        Heterotrophic yield coefficient, in[g COD/g COD]. The default is 0.625.
    f_XI_H : float, optional
        Fraction of inert COD generated in heterotrophic biomass lysis,
        in [g COD/g COD]. The default is 0.1.
    Y_PAO : float, optional
        PAO yield coefficient, in[g COD/g COD]. The default is 0.625.
    Y_PO4 : float, optional
        PP requirement (PO4 release) per PHA stored, in [g P/g COD].
        The default is 0.4.
    Y_PHA : float, optional
        PHA requirement for PP storage, in [g COD/g P]. The default is 0.2.
    f_XI_PAO : float, optional
        Fraction of inert COD generated in PAO lysis, in [g COD/g COD].
        The default is 0.1.
    Y_A : float, optional
        Autotrophic yield, in [g COD/g N]. The default is 0.24.
    f_XI_AUT : float, optional
        Fraction of inert COD generated in autotrophic biomass lysis,
        in [g COD/g COD]. The default is 0.1.
    K_h : float, optional
        Hydrolysis rate constant, in [d^(-1)]. The default is 3.0.
    eta_NO3 : float, optional
        Reduction factor for anoxic hydrolysis, dimensionless. The default is 0.6.
    eta_fe : float, optional
        Anaerobic hydrolysis reduction factor, dimensionless. The default is 0.4.
    K_O2 : float, optional
        O2 half saturation coefficient for hydrolysis, in [g O2/m^3]. The default is 0.2.
    K_NO3 : float, optional
        Nitrate half saturation coefficient for hydrolysis, in [g N/m^3].
        The default is 0.5.
    K_X : float, optional
        Slowly biodegradable substrate half saturation coefficient for hydrolysis,
        in [g COD/g COD]. The default is 0.1.
    mu_H : float, optional
        Heterotrophic maximum specific growth rate, in [d^(-1)]. The default is 6.0.
    q_fe : float, optional
        Fermentation maximum rate, in [d^(-1)]. The default is 3.0.
    eta_NO3_H : float, optional
        Reduction factor for anoxic heterotrophic growth, dimensionless.
        The default is 0.8.
    b_H : float, optional
        Lysis and decay rate constant, in [d^(-1)]. The default is 0.4.
    K_O2_H : float, optional
        O2 half saturation coefficient for heterotrophs, in [g O2/m^3].
        The default is 0.2.
    K_F : float, optional
        Fermentable substrate half saturation coefficient for heterotrophic growth,
        in [g COD/m^3]. The default is 4.0.
    K_fe : float, optional
        Fermentable substrate half saturation coefficient for fermentation,
        in [g COD/m^3]. The default is 4.0.
    K_A_H : float, optional
        VFA half saturation coefficient for heterotrophs, in [g COD/m^3].
        The default is 4.0.
    K_NO3_H : float, optional
        Nitrate half saturation coefficient for heterotrophs, in [g N/m^3].
        The default is 0.5.
    K_NH4_H : float, optional
        Ammonium (nutrient) half saturation coefficient for heterotrophs, in [g N/m^3].
        The default is 0.05.
    K_P_H : float, optional
        Phosphorus (nutrient) half saturation coefficient for heterotrophs,
        in [g P/m^3]. The default is 0.01.
    K_ALK_H : float, optional
        Alkalinity half saturation coefficient for heterotrophs, in [mol(HCO3-)/m^3].
        The default is 0.1.
    q_PHA : float, optional
        Rate constant for storage of PHA, in [d^(-1)]. The default is 3.0.
    q_PP : float, optional
        Rate constant for storage of PP, in [d^(-1)]. The default is 1.5.
    mu_PAO : float, optional
        PAO maximum specific growth rate, in [d^(-1)]. The default is 1.0.
    eta_NO3_PAO : float, optional
        Reduction factor for anoxic growth of PAO, dimensionless. The default is 0.6.
    b_PAO : float, optional
        PAO lysis rate constant, in [d^(-1)]. The default is 0.2.
    b_PP : float, optional
        PP lysis rate constant, in [d^(-1)]. The default is 0.2.
    b_PHA : float, optional
        PHA lysis rate constant, in [d^(-1)]. The default is 0.2.
    K_O2_PAO : float, optional
        O2 half saturation coefficient for PAO, in [g O2/m^3]. The default is 0.2.
    K_NO3_PAO : float, optional
        Nitrate half saturation coefficient for PAO, in [g N/m^3]. The default is 0.5.
    K_A_PAO : float, optional
        VFA half saturation coefficient for PAO, in [g COD/m^3]. The default is 4.0.
    K_NH4_PAO : float, optional
        Ammonium (nutrient) half saturation coefficient for PAO, in [g N/m^3].
        The default is 0.05.
    K_PS : float, optional
        Phosphorus half saturation coefficient for storage of PP, in [g P/m^3].
        The default is 0.2.
    K_P_PAO : float, optional
        Phosphorus (nutrient) half saturation coefficient for PAO, in [g P/m^3].
        The default is 0.01.
    K_ALK_PAO : float, optional
        Alkalinity half saturation coefficient for PAO, in [mol(HCO3-)/m^3].
        The default is 0.1.
    K_PP : float, optional
        Poly-phosphate half saturation coefficient for storage of PHA, in [g P/g COD].
        The default is 0.01.
    K_MAX : float, optional
        Maximum ratio of poly-phosphate to PAO, in [g P/g COD]. The default is 0.34.
    K_IPP : float, optional
        Inhibition coefficient for poly-phosphate storage, in [g P/g COD].
        The default is 0.02.
    K_PHA : float, optional
        PHA half saturation coefficient, in [g COD/g COD]. The default is 0.01.
    mu_AUT : float, optional
        Autotrophic maximum specific growth rate, in [d^(-1)]. The default is 1.0.
    b_AUT : float, optional
        Autotrophic decay rate, in [d^(-1)]. The default is 0.15.
    K_O2_AUT : float, optional
        O2 half saturation coefficient for autotrophs, in [g O2/m^3].
        The default is 0.5.
    K_NH4_AUT : float, optional
        Ammonium (nutrient) half saturation coefficient for autotrophs, in [g N/m^3].
        The default is 1.0.
    K_ALK_AUT : float, optional
        Alkalinity half saturation coefficient for autotrophs, in [mol(HCO3-)/m^3].
        The default is 0.5.
    K_P_AUT : float, optional
        Phosphorus (nutrient) half saturation coefficient for autotrophs, in [g P/m^3].
        The default is 0.01.
    k_PRE : float, optional
        Rate constant for phosphorus precipitation with Fe(OH)3, in [m^3/g Fe(OH)3/d].
        The default is 1.0.
    k_RED : float, optional
        Rate constant for redissoluation of FePO4, in [d^(-1)]. The default is 0.6.
    K_ALK_PRE : float, optional
        Alkalinity half saturation coefficient for P precipitation, in [mol(HCO3-)/m^3].
        The default is 0.5.
    path : str, optional
        Alternative file path for the Gujer matrix. The default is None.


    References
    ----------
    .. [1] Henze, M.; Gujer, W.; Mino, T.; Loosdrecht, M. van. Activated Sludge
        Models: ASM1, ASM2, ASM2d and ASM3; IWA task group on mathematical modelling
        for design and operation of biological wastewater treatment, Ed.; IWA
        Publishing: London, 2000.
    .. [2] Rieger, L.; Gillot, S.; Langergraber, G.; Ohtsuki, T.; Shaw, A.; Takács,
        I.; Winkler, S. Guidelines for Using Activated Sludge Models; IWA Publishing:
        London, New York, 2012; Vol. 11.
        https://doi.org/10.2166/9781780401164.

    '''
    def __new__(cls, components=None,
                iN_SI=0.01, iN_SF=0.03, iN_XI=0.02, iN_XS=0.04, iN_BM=0.07,
                iP_SI=0.0, iP_SF=0.01, iP_XI=0.01, iP_XS=0.01, iP_BM=0.02,
                iTSS_XI=0.75, iTSS_XS=0.75, iTSS_BM=0.9,
                f_SI=0.0, Y_H=0.625, f_XI_H=0.1,
                Y_PAO=0.625, Y_PO4=0.4, Y_PHA=0.2, f_XI_PAO=0.1,
                Y_A=0.24, f_XI_AUT=0.1,
                K_h=3.0, eta_NO3=0.6, eta_fe=0.4, K_O2=0.2, K_NO3=0.5, K_X=0.1,
                mu_H=6.0, q_fe=3.0, eta_NO3_H=0.8, b_H=0.4, K_O2_H=0.2, K_F=4.0,
                K_fe=4.0, K_A_H=4.0, K_NO3_H=0.5, K_NH4_H=0.05, K_P_H=0.01, K_ALK_H=0.1,
                q_PHA=3.0, q_PP=1.5, mu_PAO=1.0, eta_NO3_PAO=0.6, b_PAO=0.2, b_PP=0.2,
                b_PHA=0.2, K_O2_PAO=0.2, K_NO3_PAO=0.5, K_A_PAO=4.0, K_NH4_PAO=0.05,
                K_PS=0.2, K_P_PAO=0.01, K_ALK_PAO=0.1,
                K_PP=0.01, K_MAX=0.34, K_IPP=0.02, K_PHA=0.01,
                mu_AUT=1.0, b_AUT=0.15, K_O2_AUT=0.5, K_NH4_AUT=1.0, K_ALK_AUT=0.5, K_P_AUT=0.01,
                k_PRE=1.0, k_RED=0.6, K_ALK_PRE=0.5,
                path=None, **kwargs):
        if not path: path = data_path

        self = Processes.load_from_file(path,
                                        components=components,
                                        conserved_for=('COD', 'N', 'P', 'charge'),
                                        parameters=params,
                                        compile=False)
        if path == None:
            self.extend([p12, p14])

        self.compile()

        self._components.S_I.i_N = iN_SI
        self._components.S_F.i_N = iN_SF
        self._components.X_I.i_N = iN_XI
        self._components.X_S.i_N = iN_XS
        self._components.X_H.i_N = self._components.X_PAO.i_N = self._components.X_AUT.i_N = iN_BM
        self._components.S_I.i_P = iP_SI
        self._components.S_F.i_P = iP_SF
        self._components.X_I.i_P = iP_XI
        self._components.X_S.i_P = iP_XS
        self._components.X_H.i_P = self._components.X_PAO.i_P = self._components.X_AUT.i_P = iP_BM
        self._components.X_I.i_mass = iTSS_XI
        self._components.X_S.i_mass = iTSS_XS
        self._components.X_H.i_mass = self._components.X_PAO.i_mass = self._components.X_AUT.i_mass = iTSS_BM

        self.set_parameters(f_SI=f_SI, Y_H=Y_H, f_XI_H=f_XI_H, Y_PAO=Y_PAO, Y_PO4=Y_PO4,
                            Y_PHA=Y_PHA, f_XI_PAO=f_XI_PAO, Y_A=Y_A, f_XI_AUT=f_XI_AUT,
                            K_h=K_h, eta_NO3=eta_NO3, eta_fe=eta_fe, K_O2=K_O2,
                            K_NO3=K_NO3, K_X=K_X,
                            mu_H=mu_H, q_fe=q_fe, eta_NO3_H=eta_NO3_H, b_H=b_H,
                            K_O2_H=K_O2_H, K_F=K_F, K_fe=K_fe, K_A_H=K_A_H, K_NO3_H=K_NO3_H,
                            K_NH4_H=K_NH4_H, K_P_H=K_P_H, K_ALK_H=K_ALK_H*12,
                            q_PHA=q_PHA, q_PP=q_PP, mu_PAO=mu_PAO, eta_NO3_PAO=eta_NO3_PAO,
                            b_PAO=b_PAO, b_PP=b_PP, b_PHA=b_PHA, K_O2_PAO=K_O2_PAO,
                            K_NO3_PAO=K_NO3_PAO, K_A_PAO=K_A_PAO, K_NH4_PAO=K_NH4_PAO,
                            K_PS=K_PS, K_P_PAO=K_P_PAO, K_ALK_PAO=K_ALK_PAO*12, K_PP=K_PP,
                            K_MAX=K_MAX, K_IPP=K_IPP, K_PHA=K_PHA,
                            mu_AUT=mu_AUT, b_AUT=b_AUT, K_O2_AUT=K_O2_AUT, K_NH4_AUT=K_NH4_AUT,
                            K_ALK_AUT=K_ALK_AUT*12, K_P_AUT=K_P_AUT,
                            k_PRE=k_PRE, k_RED=k_RED, K_ALK_PRE=K_ALK_PRE*12,
                            **kwargs)
        return self


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


# %%

import numpy as np
from .. import Construction
from ._decay import Decay
from ._sludge_separator import SludgeSeparator
from ..utils import load_data, data_path

__all__ = ('SedimentationTank',)

data_path += 'sanunit_data/_sedimentation_tank.tsv'


class SedimentationTank(SludgeSeparator, Decay):
    '''
    Sedimentation of wastes into liquid and solid phases based on Trimmer et al. [1]_

    Parameters
    ----------
    ins : WasteStream
        Waste for treatment.
    outs : WasteStream
        Liquid, settled solids, fugitive CH4, and fugitive N2O.
    degraded_components : tuple
        IDs of components that will degrade (simulated by first-order decay).
    if_N2O_emission : bool
        If consider N2O emission from N degradation the process.

    Examples
    --------
    `bwaise systems <https://github.com/QSD-Group/EXPOsan/blob/main/exposan/bwaise/systems.py>`_

    References
    ----------
    .. [1] Trimmer et al., Navigating Multidimensional Social–Ecological System
        Trade-Offs across Sanitation Alternatives in an Urban Informal Settlement.
        Environ. Sci. Technol. 2020, 54 (19), 12641–12653.
        https://doi.org/10.1021/acs.est.0c03296.

    See Also
    --------
    :ref:`qsdsan.sanunits.Decay <sanunits_Decay>`
    '''

    def __init__(self, ID='', ins=None, outs=(),thermo=None, init_with='WasteStream',
                 split=None, settled_frac=None,
                 degraded_components=('OtherSS',), if_N2O_emission=False, **kwargs):

        SludgeSeparator.__init__(self, ID, ins, outs, thermo, init_with,
                                 split, settled_frac, F_BM_default=1)
        self.degraded_components = tuple(degraded_components)
        self.if_N2O_emission = if_N2O_emission

        data = load_data(path=data_path)
        for para in data.index:
            value = float(data.loc[para]['expected'])
            setattr(self, '_'+para, value)
        del data

        for attr, value in kwargs.items():
            setattr(self, attr, value)

    _N_ins = 1
    _N_outs = 4

    def _run(self):
        waste = self.ins[0]
        liq, sol, CH4, N2O = self.outs
        CH4.phase = N2O.phase = 'g'

        # Retention in the settled solids
        SludgeSeparator._run(self)

        # COD degradation in settled solids
        COD_loss = self.first_order_decay(k=self.decay_k_COD,
                                          t=self.tau/365,
                                          max_decay=self.COD_max_decay)
        tot_COD_kg = sol._COD * sol.F_vol / 1e3
        sol.imass[self.degraded_components] *= 1 - COD_loss

        # Adjust total mass of of the settled solids by changing water content
        liq, sol = self._adjust_solid_water(waste, liq, sol)

        COD_loss_kg = tot_COD_kg * COD_loss
        CH4.imass['CH4'] = COD_loss_kg * self.max_CH4_emission * self.MCF_decay
        sol._COD = tot_COD_kg*(1-COD_loss)/sol.F_vol*1e3

        # N degradation
        if self.if_N2O_emission:
            N_loss = self.first_order_decay(k=self.decay_k_N,
                                            t=self.tau/365,
                                            max_decay=self.N_max_decay)
            N_loss_tot = N_loss*sol.TN/1e3*sol.F_vol
            NH3_rmd, NonNH3_rmd = \
                self.allocate_N_removal(N_loss_tot, sol.imass['NH3'])
            sol.imass ['NH3'] -=  NH3_rmd
            sol.imass['NonNH3'] -= NonNH3_rmd
            N2O.imass['N2O'] = N_loss_tot*self.N2O_EF_decay*44/28
        else:
            N2O.empty()

    _units = {
        'Single tank volume': 'm3',
        'Single tank height': 'm',
        'Single tank width': 'm',
        'Single tank length': 'm',
        'Single roof area': 'm2'
        }

    def _design(self):
        design = self.design_results
        #!!! Why isn't tau used?
        design['Tank number'] = N = self.N_tank
        design['Single tank volume'] = V_single = self.tank_V
        L2W = self.tank_L_to_W
        W2H = self.tank_W_to_H
        design['Single tank height'] = H = (V_single/(L2W*(W2H**2)))**(1/3)
        design['Single tank width'] = W = H * W2H
        design['Single tank length'] = L = W * L2W
        design['Single roof area'] = N*L*W/(np.cos(self.roof_slope/180*np.pi))
        side_area = N*2*(L*H + W*H)

        # Concrete
        thick = self.concrete_thickness
        side_concrete = N*thick*(L*W+2*W*H+2*L*H)
        column_concrete = N*(thick**2)*H*self.column_per_side*2

        self.construction = (
            Construction(item='Concrete', quantity=side_concrete+column_concrete, quantity_unit='m3'),
            Construction(item='Excavation', quantity=design['Single roof area']+side_area, quantity_unit='m3'),
            )
        self.add_construction()

    @property
    def tau(self):
        '''[float] Residence time, [d].'''
        return self._tau
    @tau.setter
    def tau(self, i):
        self._tau = i

    @property
    def tank_V(self):
        '''[float] Volume of the sedimentation tank.'''
        return self._tank_V
    @tank_V.setter
    def tank_V(self, i):
        self._tank_V = i

    @property
    def tank_L_to_W(self):
        '''[float] Length-to-width ratio of the sedimentation tank.'''
        return self._tank_L_to_W
    @tank_L_to_W.setter
    def tank_L_to_W(self, i):
        self._tank_L_to_W = i

    @property
    def tank_W_to_H(self):
        '''[float] Width-to-height ratio of the sedimentation tank.'''
        return self._tank_W_to_H
    @tank_W_to_H.setter
    def tank_W_to_H(self, i):
        self._tank_W_to_H = i

    @property
    def N_tank(self):
        '''[int] Number of sedimentation tanks, float will be converted to the smallest integer.'''
        return self._N_tank
    @N_tank.setter
    def N_tank(self, i):
        self._N_tank = int(np.ceil(i))

    @property
    def column_per_side(self):
        '''[int] Number of columns per side of sedimentation tanks, float will be converted to the smallest integer.'''
        return self._column_per_side
    @column_per_side.setter
    def column_per_side(self, i):
        self._column_per_side = int(np.ceil(i))

    @property
    def concrete_thickness(self):
        '''[float] Thickness of the concrete wall.'''
        return self._concrete_thickness
    @concrete_thickness.setter
    def concrete_thickness(self, i):
        self._concrete_thickness = i

    @property
    def roof_slope(self):
        '''[float] Slope of the tank roof, [°].'''
        return self._roof_slope
    @roof_slope.setter
    def roof_slope(self, i):
        self._roof_slope = i

    @property
    def roof_unit_mass(self):
        '''[float] Unit mass of the tank roof, [kg/m2].'''
        return self._roof_unit_mass
    @roof_unit_mass.setter
    def roof_unit_mass(self, i):
        self._roof_unit_mass = i


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

With contributions from:
    Joy Cheung <joycheung1994@gmail.com>
    Yoel Rene Cortés-Peña <yoelcortes@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


# %%

__all__ = ('get_correlations', 'define_inputs', 'generate_samples',
           'morris_analysis', 'morris_till_convergence', 'sobol_analysis',
           'plot_uncertainties', 'plot_correlations',
           'plot_morris_results', 'plot_morris_convergence', 'plot_sobol_results')

from collections.abc import Iterable
import numpy as np
import pandas as pd
import seaborn as sns
import biosteam as bst
from warnings import warn
from matplotlib import pyplot as plt
from SALib.sample import (
    morris as morris_sampler,
    fast_sampler,
    latin as rbd_sampler,
    saltelli as sobol_sampler)
from SALib.analyze import morris, fast, rbd_fast, sobol
from SALib.plotting import morris as sa_plt_morris
from biosteam.plots import plot_spearman
from .utils import time_printer

isinstance = isinstance
getattr = getattr
var_indices = bst.evaluation._model.var_indices
indices_to_multiindex = bst.evaluation._model.indices_to_multiindex


# %%

# =============================================================================
# Utility functions
# =============================================================================

def _update_input(input_val, default_val):
    if input_val is None:
        return default_val
    else:
        try:
            iter(input_val)
            if len(input_val)==0: # empty iterable
                return default_val
            return input_val if not isinstance(input_val, str) else (input_val,)
        except:
            return (input_val,)


def _update_nan(df, nan_policy, legit=('propagate', 'raise', 'omit')):
    if not nan_policy in legit:
        raise ValueError(f'nan_policy can only be in {legit}, not "{nan_policy}".')
    if nan_policy == 'propagate':
        return 'nan'
    elif nan_policy == 'raise':
        raise ValueError('"NaN" values in inputs, cannot run analysis.')
    elif nan_policy == 'omit':
        return df.dropna()
    elif nan_policy == 'fill_mean':
        return df.fillna(df.dropna().mean())
    # Shouldn't get to this step
    else:
        return df

def _update_df_names(df, columns=True, index=True):
    new_df = df.copy()

    if columns and not new_df.columns.empty:
        try:
            iter(new_df.columns)
            new_df.columns = [i[-1].split(' [')[0] for i in new_df.columns]
        except: pass

    if index and not new_df.index.empty:
        try:
            iter(new_df.index)
            new_df.index = [i[-1].split(' [')[0] for i in new_df.index]
        except: pass

    return new_df


def _save_fig_return(fig, ax, file, close_fig):
    if file:
        fig.savefig(file, dpi=300)

    if close_fig:
        plt.close()

    return fig, ax


# %%

# =============================================================================
# Correlations
# =============================================================================

def get_correlations(model, input_x=None, input_y=None,
                     kind='Spearman', nan_policy='propagate', file='',
                     **kwargs):
    '''
    Get correlation coefficients between two inputs using ``scipy``.

    Parameters
    ----------
    model : :class:`biosteam.Model`
        Uncertainty model with defined paramters and metrics.
    input_x : :class:`biosteam.Parameter` or :class:`biosteam.Metric`
        The first set of input, can be single values or an iterable,
        will be defaulted to all model parameters if not provided.
    input_y : :class:`biosteam.Parameter` or :class:`biosteam.Metric`
        The second set of input, can be single values or an iterable,
        will be defaulted to all model parameters if not provided.
    kind : str
        Can be "Spearman" for Spearman's rho, "Pearson" for Pearson's r,
        "Kendall" for Kendall's tau, or "KS" for Kolmogorov–Smirnov's D.
    nan_policy : str
        - "propagate": returns nan.
        - "raise": raise an error.
        - "omit": drop the pair from analysis.
    file : str
        If provided, the results will be saved as an Excel file.
    kwargs
        Other kwargs that will be passed to ``scipy``.

    Returns
    -------
    Two :class:`pandas.DataFrame` containing the test statistics and p-values.

    Examples
    --------
    `QSDsan.stats <https://qsdsan.readthedocs.io/en/latest/stats.html>`_

    See Also
    --------
    :func:`scipy.stats.spearmanr`

    :func:`scipy.stats.pearsonr`

    :func:`scipy.stats.kendalltau`

    :func:`scipy.stats.kstest`
    '''

    if input_x and not isinstance(input_x, Iterable): input_x = (input_x,)
    if input_y and not isinstance(input_y, Iterable): input_y = (input_y,)
    if nan_policy not in ('propagate', 'raise', 'omit'):
        raise ValueError(f'nan_policy can only be in ("omit", "propagate", "raise"), ' \
                         f'not "{nan_policy}".')

    name = kind.lower()
    if name == 'spearman':
        correlation = model.spearman_r
        sheet_name = 'rho'
    elif name == 'pearson':
        correlation = model.pearson_r
        sheet_name = 'r'
    elif name == 'kendall':
        correlation = model.kendall_tau
        sheet_name = 'tau'
    elif name == 'ks':
        correlation = model.kolmogorov_smirnov_d
        sheet_name = 'D'
    else:
        raise ValueError('kind can only be "Spearman", "Pearson", '
                        f'"Kendall", or "KS", not "{kind}".')
    r_df, p_df = dfs = correlation(input_x, input_y, nan_policy+' nan', **kwargs)
    for df in dfs:
        df.index.names = ('Element', 'Input x')
        df.columns.names = ('Element', 'Input y')

    if file:
        with pd.ExcelWriter(file) as writer:
            r_df.to_excel(writer, sheet_name=sheet_name)
            p_df.to_excel(writer, sheet_name='p-value')
    return r_df, p_df


# %%

# =============================================================================
# SALib modules
# =============================================================================

def define_inputs(model):
    '''
    Define the model inputs (referred to as "problem") to be used for sampling by ``SALib``.

    Parameters
    ----------
    model : :class:`biosteam.Model`
        Uncertainty model with defined paramters and metrics.

    Returns
    -------
    inputs : dict
        A dict containing model inputs for sampling by ``SALib``.

    Examples
    --------
    `QSDsan.stats <https://qsdsan.readthedocs.io/en/latest/stats.html>`_

    See Also
    --------
    `SALib Basics <https://salib.readthedocs.io/en/latest/basics.html#an-example>`_
    '''
    return model.problem()


# TODO: incorporate sampling from different distributions
def generate_samples(inputs, kind, N, seed=None, **kwargs):
    '''
    Generate samples for sensitivity analysis using ``SALib``.

    Parameters
    ----------
    model : :class:`biosteam.Model`
        Uncertainty model with defined paramters and metrics.
    inputs : dict
        A dict generated by :func:`qsdsan.sensitivity.define_inputs` to be used for ``SALib``,
        keys should include "num_vars", "names", and "bounds".
    kind : str
        Can be "FAST" for Fourier amplitude sensitivity test (FAST),
        "RBD" for random balance design FAST (latin hypercube sampling [LHS]),
        "Morris" for Morris One-at-A-Time (OAT),
        or "Sobol" for Sobol analysis (Saltelli sampling).
    N : int
        The number of samples or trajectories (Morris).
    seed : int
        Seed to generate random samples.
    kwargs
        Other kwargs that will be passed to ``SALib``.

    Returns
    -------
    samples: array
        Samples to be used for the indicated sensitivies analyses.

    Examples
    --------
    `QSDsan.stats <https://qsdsan.readthedocs.io/en/latest/stats.html>`_

    See Also
    --------
    ``SALib`` `API <https://salib.readthedocs.io/en/latest/api.html>`_
    '''

    lower = kind.lower()
    if lower == 'morris':
        return morris_sampler.sample(inputs, N=N, seed=seed, **kwargs)
    elif lower in ('fast', 'efast'):
        return fast_sampler.sample(inputs, N=N, seed=seed, **kwargs)
    elif lower == 'rbd':
        return rbd_sampler.sample(inputs, N=N, seed=seed, **kwargs)
    elif lower == 'sobol':
        return sobol_sampler.sample(inputs, N=N, seed=seed, **kwargs)
    else:
        raise ValueError('kind can only be "FAST", "RBD", "Morris", or "Sobol", ' \
                         f'not "{kind}".')

# =============================================================================
# Morris
# =============================================================================

@time_printer
def morris_analysis(model, inputs, metrics=None, nan_policy='propagate',
                    conf_level=0.95, print_to_console=False,
                    print_time=False, file='', **kwargs):
    '''
    Run Morris sensitivity analysis using ``SALib``.

    Parameters
    ----------
    model : :class:`biosteam.Model`
        Uncertainty model with defined paramters and metrics.
    inputs : dict
        A dict generated by :func:`qsdsan.sensitivity.define_inputs` to be used for ``SALib``,
        keys should include "num_vars", "names", and "bounds".
    metrics : :class:`biosteam.Metric`
        Metrics to be included for Morris analysis, must be a subset of
        the metrics of the model to be analyzed.
        (i.e., included in the `metrics` attribute of the given model).
        If None is provided, all metrics in the model will be included.
    nan_policy : str
        - "propagate": returns nan.
        - "raise": raise an error.
        - "fill_mean": fill nan with mean of the results.
    conf_level : float
        Confidence level of results.
    print_to_console : bool
        Whether to show results in the console.
    print_time : bool
        Whether to show simulation time in the console.
    file : str
        If provided, the results will be saved as an Excel file.
    kwargs
        Other kwargs that will be passed to ``SALib``.

    Returns
    -------
    morris_dct : dict
        A dict of Morris analysis results.

    Examples
    --------
    `QSDsan.stats <https://qsdsan.readthedocs.io/en/latest/stats.html>`_

    See Also
    --------
    ``SALib`` `API <https://salib.readthedocs.io/en/latest/api.html>`_
    '''

    morris_dct = {}
    model = model.copy()
    metrics = _update_input(metrics, model.metrics)
    model.metrics = metrics
    table = model.table.astype('float64')
    table = _update_nan(table, nan_policy, legit=('propagate', 'raise', 'fill_mean'))

    if isinstance(table, str):
        table = model.table.astype('float64')

    param_val = table.iloc[:, :len(model.get_parameters())]
    metric_val = pd.concat([table[metric.index] for metric in metrics], axis=1)

    for metric in metrics:
        results = metric_val[metric.index]
        si = morris.analyze(inputs, param_val.to_numpy(), results.to_numpy(),
                            conf_level=conf_level, print_to_console=print_to_console,
                            **kwargs)
        morris_dct[metric.name] = si.to_df()

    if file:
        writer = pd.ExcelWriter(file)
        for name, si_df in morris_dct.items():
            si_df.to_excel(writer, sheet_name=name)
        writer.save()

    return morris_dct


@time_printer
def morris_till_convergence(model, inputs, metrics=None,
                            N_max=20, seed=None, threshold=0.1,
                            nan_policy='propagate',
                            conf_level=0.95, print_to_console=False,
                            print_time=False, file='', **kwargs):
    '''
    Run Morris analysis from N=2 to N=N_max until the results converge
    (i.e., mu_star_conf/mu_star_max < threshold for all parameters,
     where as mu_star_max is the maximum :math:`{\mu^*}` value for a certain metric,
     and this should be satisfied for all metrics).

    Parameters
    ----------
    model : :class:`biosteam.Model`
        Uncertainty model with defined paramters and metrics.
    inputs : dict
        A dict generated by :func:`qsdsan.sensitivity.define_inputs` to be used for ``SALib``,
        keys should include "num_vars", "names", and "bounds".
    metrics : :class:`biosteam.Metric`
        Metrics to be included for Morris analysis, must be a subset of
        the metrics of the model to be analyzed.
        (i.e., included in the `metrics` attribute of the given model).
        If None is provided, all metrics in the model will be included.
    N_max : int
        Maximum number of trajectories to be considered.
    seed : int
        Seed to generate random samples.
    threshold : float
        Threshold for the convergence.
    nan_policy : str
        - "propagate": returns nan.
        - "raise": raise an error.
        - "fill_mean": fill nan with mean of the results.
    conf_level : float
        Confidence level of results.
    print_to_console : bool
        Whether to show results in the console.
    print_time : bool
        Whether to show simulation time in the console.
    file : str
        If provided, the results will be saved as an Excel file.
    kwargs
        Other kwargs that will be passed to ``SALib``.

    Examples
    --------
    `QSDsan.stats <https://qsdsan.readthedocs.io/en/latest/stats.html>`_

    See Also
    --------
    :func:`qsdsan.stats.morris_analysis`
    '''

    num_levels = kwargs['num_levels'] if 'num_levels' in kwargs.keys() else 4
    kwargs = {i:kwargs[i] for i in kwargs.keys() if i!='num_levels'}
    samples = generate_samples(inputs=inputs, kind='Morris', N=N_max,
                               seed=seed, num_levels=num_levels)
    model.load_samples(samples)

    param_num = len(model.get_parameters())
    cum_model = model.copy()
    cum_model.load_samples(samples[0: 2*(param_num+1)])
    cum_model.evaluate()
    cum_dct = dict(mu_star={}, mu_star_conf={})
    metrics = _update_input(metrics, model.metrics)
    temp_dct = morris_analysis(model=cum_model, inputs=inputs, metrics=metrics,
                               nan_policy=nan_policy, conf_level=conf_level,
                               print_to_console=print_to_console, **kwargs)
    for m in metrics:
        for idx in ('mu_star', 'mu_star_conf'):
            data0 = getattr(temp_dct[m.name], idx)
            df = pd.DataFrame(columns=data0.index, index=(2,))
            df.index.name = idx
            df.loc[2] = data0.copy()
            cum_dct[idx][m.name] = df

    for n in range(2, N_max):
        temp_model = model.copy()
        temp_model.load_samples(samples[n*(param_num+1): (n+1)*(param_num+1)])
        temp_model.evaluate()
        cum_model.table = pd.concat((cum_model.table, temp_model.table))

        temp_dct = morris_analysis(model=cum_model, inputs=inputs, metrics=metrics,
                                   nan_policy=nan_policy, conf_level=conf_level,
                                   print_to_console=print_to_console, **kwargs)
        all_converged = True
        for m in metrics:
            mu_star = temp_dct[m.name].mu_star
            mu_star_conf = temp_dct[m.name].mu_star_conf
            cum_dct['mu_star'][m.name].loc[n+1] = mu_star
            cum_dct['mu_star_conf'][m.name].loc[n+1] = mu_star_conf

            converged = False if (mu_star_conf/mu_star.max()>threshold).any() else True
            all_converged = all_converged & converged

        if all_converged:
            print(f'mu_star converges at {n} trajectories.')
            break
        elif n == N_max-1:
            print(f'mu_star has not converged with {n} trajectories.')

    if file:
        writer = pd.ExcelWriter(file)
        for m in metrics:
            cum_dct['mu_star'][m.name].to_excel(writer, sheet_name=m.name)
            cum_dct['mu_star_conf'][m.name].to_excel(
                writer, sheet_name=m.name, startrow=N_max)
        writer.save()

    return cum_dct


# =============================================================================
# (e)FAST and RBD-FAST
# =============================================================================

@time_printer
def fast_analysis(model, inputs, kind, metrics=None, nan_policy='propagate',
                  conf_level=0.95, print_to_console=False, print_time=False,
                  file='', **kwargs):
    '''
    Run Fourier amplitude sensitivity test (Saltelli's extended FAST) or
    random balance design (RBD) FAST using ``SALib``.

    Parameters
    ----------
    model : :class:`biosteam.Model`
        Uncertainty model with defined paramters and metrics.
    inputs : dict
        A dict generated by :func:`qsdsan.sensitivity.define_inputs` to be used for ``SALib``,
        keys should include "num_vars", "names", and "bounds".
    kind : str
        Either "FAST" or "RBD".
    metrics : :class:`biosteam.Metric`
        Metrics to be included for Morris analysis, must be a subset of
        the metrics of the model to be analyzed.
        (i.e., included in the `metrics` attribute of the given model).
        If None is provided, all metrics in the model will be included.
    nan_policy : str
        - "propagate": returns nan.
        - "raise": raise an error.
        - "fill_mean": fill nan with mean of the results.
    conf_level : float
        Confidence level of results.
    print_to_console : bool
        Whether to show results in the console.
    print_time : bool
        Whether to show simulation time in the console.
    file : str
        If provided, the results will be saved as an Excel file.
    kwargs
        Other kwargs that will be passed to ``SALib``.

    Returns
    -------
    fast_dct : dict
        A dict of FAST analysis results.

    Examples
    --------
    `QSDsan.stats <https://qsdsan.readthedocs.io/en/latest/stats.html>`_

    See Also
    --------
    ``SALib`` `API <https://salib.readthedocs.io/en/latest/api.html>`_
    '''

    fast_dct = {}
    model = model.copy()
    metrics = _update_input(metrics, model.metrics)
    model.metrics = metrics
    table = model.table.astype('float64')
    table = _update_nan(table, nan_policy, legit=('propagate', 'raise', 'fill_mean'))

    if isinstance(table, str):
        table = model.table.astype('float64')

    param_val = table.iloc[:, :len(model.get_parameters())]
    metric_val = pd.concat([table[metric.index] for metric in metrics], axis=1)

    if kind.lower() in ('fast', 'efast'):
        for metric in metrics:
            results = metric_val[metric.index]
            si = fast.analyze(inputs, results.to_numpy(), conf_level=conf_level,
                              print_to_console=print_to_console, **kwargs)
            fast_dct[metric.name] = si.to_df()

    elif kind.lower() == 'rbd':
        for metric in metrics:
            results = metric_val[metric.index]
            si = rbd_fast.analyze(inputs, param_val.to_numpy(), results.to_numpy(),
                                  conf_level=conf_level, print_to_console=print_to_console,
                                  **kwargs)
            fast_dct[metric.name] = si.to_df()

    else:
        raise ValueError(f'kind can only be "FAST" or "RBD", not "{kind}".')

    if file:
        writer = pd.ExcelWriter(file)
        for name, si_df in fast_dct.items():
            si_df.to_excel(writer, sheet_name=name)
        writer.save()

    return fast_dct


# =============================================================================
# Sobol
# =============================================================================

@time_printer
def sobol_analysis(model, inputs, metrics=None, nan_policy='propagate',
                   calc_second_order=True, conf_level=0.95, print_to_console=False,
                   print_time=False, file='', **kwargs):
    '''
    Run Sobol sensitivity analysis using ``SALib``.

    Parameters
    ----------
    model : :class:`biosteam.Model`
        Uncertainty model with defined paramters and metrics.
    inputs : dict
        A dict generated by :func:`qsdsan.sensitivity.define_inputs` to be used for ``SALib``,
        keys should include "num_vars", "names", and "bounds".
    metrics : :class:`biosteam.Metric`
        Metrics to be included for Morris analysis, must be a subset of
        the metrics of the model to be analyzed.
        (i.e., included in the `metrics` attribute of the given model).
        If None is provided, all metrics in the model will be included.
    nan_policy : str
        - "propagate": returns nan.
        - "raise": raise an error.
        - "fill_mean": fill nan with mean of the results.
    calc_second_order : bool
        Whether to calculate second-order interaction effects.
    conf_level : float
        Confidence level of results.
    print_to_console : bool
        Whether to show results in the console.
    print_time : bool
        Whether to show simulation time in the console.
    file : str
        If provided, the results will be saved as an Excel file.
    kwargs
        Other kwargs that will be passed to ``SALib``.

    Returns
    -------
    si_dct : dict
        A dict of Sobol analysis results.

    Examples
    --------
    `QSDsan.stats <https://qsdsan.readthedocs.io/en/latest/stats.html>`_

    See Also
    --------
    ``SALib`` `API <https://salib.readthedocs.io/en/latest/api.html>`_
    '''

    sobol_dct = {}
    metrics = _update_input(metrics, model.metrics)
    model = model.copy()
    model.metrics = metrics
    table = model.table.astype('float64')

    df = pd.concat([table[metric.index] for metric in metrics], axis=1)
    results = _update_nan(df, nan_policy, legit=('propagate', 'raise', 'fill_mean'))
    if isinstance(results, str):
        results = df

    for metric in metrics:
        result = results[metric.index]
        si = sobol.analyze(inputs, result.to_numpy(),
                           calc_second_order=calc_second_order,
                           conf_level=conf_level, print_to_console=print_to_console,
                           **kwargs)
        sobol_dct[metric.name] = dict(zip(('ST', 'S1', 'S2'), si.to_df()))

    if file:
        writer = pd.ExcelWriter(file)
        for name, si_df in sobol_dct.items():
            n_row = 0
            for df in si_df.values():
                df.to_excel(writer, sheet_name=name, startrow=n_row)
                n_row += len(df.index) + 2 + len(df.columns.names)
        writer.save()

    return sobol_dct


# %%

# =============================================================================
# Plot uncertainty analysis results
# =============================================================================

def plot_uncertainties(model, x_axis=(), y_axis=(), kind='box',
                       file='', close_fig=True, center_kws={}, margin_kws={}):
    '''
    Visualize uncertainty analysis results as one of the following depending on inputs:

    +---------------------------------+-----------------------------------------------+
    | input                           | returned plot                                 |
    +----------+----------+-----------+-------------+----------------+----------------+
    | x_axis   | y_axis   | kind      | orientation | center         | margin         |
    +==========+==========+===========+=============+================+================+
    | single   | None     | box       | 1D          | box            | N/A            |
    | or       |          +-----------+ horizontal  +----------------+                +
    | iterable |          | hist      |             | histogram      |                |
    |          |          +-----------+             +----------------+                +
    |          |          | kde       |             | kernel density |                |
    +----------+----------+-----------+-------------+----------------+----------------+
    | None     | single   | box       | 1D          | box            | N/A            |
    |          | or       +-----------+ vertical    +----------------+                +
    |          | iterable | hist      |             | histogram      |                |
    |          |          +-----------+             +----------------+                +
    |          |          | kde       |             | kernel density |                |
    +----------+----------+-----------+-------------+----------------+----------------+
    | single   | single   | hist-box  | 2D          | histogram      | box            |
    |          |          +-----------+             +                +----------------+
    |          |          | hist-kde  |             |                | kernel density |
    |          |          +-----------+             +                +----------------+
    |          |          | hist-hist |             |                | histogram      |
    |          |          +-----------+             +----------------+----------------+
    |          |          | kde-box   |             | kernel density | box            |
    |          |          +-----------+             +                +----------------+
    |          |          | kde-hist  |             |                | histogram      |
    |          |          +-----------+             +                +----------------+
    |          |          | kde-kde   |             |                | kernel density |
    +----------+----------+-----------+-------------+----------------+----------------+

    .. note::
        When both x_axis and y_axis are not None (i.e., the figure is 2D),
        for clarity reasons, this function requires len(x_axis) and len(y_axis) to be both 1.

        If wanted, it is possible to use colors (``hue`` in ``seaborn``) to
        differentiated the different parameters or metrics in either x_axis or
        y_axis. To achieve this, please refer to the documentation of ``seaborn``.

        Similarly, there are other potential combinations of the center and
        margin plots that are supported by ``seaborn`` but not included here.

    Parameters
    ----------
    model : :class:`biosteam.Model`
        The model with uncertainty analysis (in <:class:`Model`.table>) results for plotting.
    x_axis : :class:`biosteam.Parameter`, :class:`biosteam.Metric` or iterable
        What to plot on the x-axis, can be parameters or metrics of the model,
        default to all model metrics included in the model result table
        if neither x nor y is provided.
    y_axis : :class:`biosteam.Parameter`, :class:`biosteam.Metric` or iterable
        What to plot on the y-axis, can be parameters or metrics of the model,
        default to None.
    kind : str
        What kind of plot to be returned, refer to the summary table for valid inputs.
    file : str
        If provided, the generated figure will be saved as a png file.
    close_fig : bool
        Whether to close the figure
        (if not close, new figure will be overlaid on the current figure).
    center_kws : dict
        Will be passed to ``seaborn`` for the center plot.
    margin_kws : dict
        Will be passed to ``seaborn`` for the margin plots.

    Returns
    -------
    figure : :class:`matplotlib.figure.Figure`
        The generated figure.
    axis : :class:`matplotlib.axes._subplots.AxesSubplot` or iterable
        The generated figure axis (or axes for 2D figure).

    Examples
    --------
    `QSDsan.stats <https://qsdsan.readthedocs.io/en/latest/stats.html>`_

    See Also
    --------
    :func:`seaborn.jointplot` `docs <https://seaborn.pydata.org/generated/seaborn.jointplot.html>`_
    '''

    kind_lower = kind.lower()
    table = model.table.astype('float64')
    df = _update_df_names(table)

    twoD = False
    x_df = y_df = None
    sns_df = pd.DataFrame(columns=('x_group', 'x_data', 'y_group', 'y_data'))

    dfs = []
    if not y_axis: # no data provided or only x, 1D, horizontal
        x_axis = _update_input(x_axis, model.metrics)
        x_df = df[[i.name for i in x_axis]]

        for x in x_df.columns:
            temp_df = sns_df.copy()
            temp_df['x_data'] = x_df[x]
            temp_df['x_group'] = x
            dfs.append(temp_df)

        sns_df = pd.concat(dfs)

    elif not x_axis: # only y, 1D, vertical
        y_df = df[[i.name for i in y_axis]]

        for y in y_df.columns:
            temp_df = sns_df.copy()
            temp_df['y_data'] = y_df[y]
            temp_df['y_group'] = y
            dfs.append(temp_df)

        sns_df = pd.concat(dfs)

    else: # x and y, 2D
        len_x = len_y = 1
        try:
            len_x = len(x_axis)
        except:
            try:
                len_y = len(y_axis)
            except:
                pass
        if not (len_x==1 and len_y==1):
            raise ValueError('Only single input allowed when both x_axis and y_axis are provided.')

        twoD = True
        x_name, y_name = x_axis.name, y_axis.name
        sns_df = df[[x_name, y_name]]

    sns.set_theme(style='ticks')
    if not twoD: # 1D plot
        if kind_lower == 'box':
            center_kws.setdefault('dodge', False)

            if x_df is not None: # horizontal box plot
                ax = sns.boxplot(data=sns_df, x='x_group', y='x_data', hue='x_group',
                                 **center_kws)
                xlabel = ''
                ylabel = 'Values'
            else:
                ax = sns.boxplot(data=sns_df, x='y_data', y='y_group', hue='y_group',
                                 **center_kws)
                xlabel = 'Values'
                ylabel = ''

            ax.set(xlabel=xlabel, ylabel=ylabel)

        elif kind_lower == 'hist':
            if x_df is not None: # horizontal hist plot
                ax = sns.histplot(data=sns_df, x='x_data', hue='x_group', **center_kws)
                ax.set_xlabel('Values')

            else: # vertical hist plot
                ax = sns.histplot(data=sns_df, y='y_data', hue='y_group', **center_kws)
                ax.set_ylabel('Values')

        elif kind_lower == 'kde':
            if x_df is not None: # horizontal kde plot
                ax = sns.kdeplot(data=sns_df, x='x_data', hue='x_group', **center_kws)
            else: # vertical kde plot
                ax = sns.kdeplot(data=sns_df, y='y_data', hue='y_group', **center_kws)

        else:
            raise ValueError('kind can only be "box", "kde", "hist", or "hist-kde", ' \
                             f'for 1D plot, not "{kind}".')

        ax.get_legend().set_title('')
        return _save_fig_return(ax.figure, ax, file, close_fig)

    else: # 2D plot
        g = sns.JointGrid(data=sns_df, x=x_name, y=y_name)

        kind_split = kind_lower.split('-')
        if kind_split[0] in ('hist', 'kde'):
            func = getattr(sns, f'{kind_split[0]}plot')
            g.plot_joint(func, **center_kws)
        else:
            raise ValueError(f'The provided kind "{kind}" is not valid for 2D plot.')

        if kind_split[1] in ('box', 'hist', 'kde'):
            func = getattr(sns, f'{kind_split[1]}plot')
            g.plot_marginals(func, **margin_kws)
        else:
            raise ValueError(f'The provided kind "{kind}" is not valid for 2D plot.')

        return _save_fig_return(g.fig, g.fig.axes, file, close_fig)


# =============================================================================
# Plot correlations
# =============================================================================

def _plot_corr_tornado(corr_df, top):
    fig, ax = plot_spearman(corr_df.iloc[:,0], top=top)

    ax.set_xlabel(corr_df.columns[0])

    for ax in fig.axes:
        for key in ax.spines.keys():
            ax.spines[key].set(color='k', linewidth=0.5, visible=True)
            ax.grid(False)

    return fig, ax


def _plot_corr_bubble(corr_df, ratio, **kwargs):
    sns.set_theme(style="whitegrid")

    margin_x = kwargs['margin_x'] if 'margin_x' in kwargs.keys() else 0.1/ratio
    margin_y = kwargs['margin_y'] if 'margin_y' in kwargs.keys() else 0.1
    kwargs = {i: kwargs[i] for i in kwargs.keys() if 'margin' not in i}

    keys = ('height', 'palette', 'hue_norm', 'sizes', 'size_norm', 'edgecolor')
    values = (5+ratio, 'vlag', (-1, 1), (0, 1000), (0, 2), '0.5')

    for num, k in enumerate(keys):
        kwargs.setdefault(keys[num], values[num])

    g = sns.relplot(data=corr_df, x='metric', y='parameter',
                    hue='correlation', size='size',  **kwargs)

    g.set(xlabel='', ylabel='', aspect=1)
    g.ax.margins(x=margin_x, y=margin_y)

    for label in g.ax.get_xticklabels():
        label.set_rotation(90)

    for artist in g.legend.legendHandles:
        artist.set_edgecolor('0.5')

    for key in g.ax.spines.keys():
        g.ax.spines[key].set(color='k', linewidth=0.5, visible=True)

    g.ax.grid(True, which='major', color='k',linestyle='--', linewidth=0.7)
    g.tight_layout()

    return g


def plot_correlations(result_df, parameters=(), metrics=(), top=None,
                      file='', close_fig=True, **kwargs):
    '''
    Visualize the correlations between model parameters and metric results
    as tornado (single metric) or bubble plots (multiple metrics).

    Parameters
    ----------
    result_df : :class:`pandas.DataFrame`
        Result table generated by :func:`get_correlations` containing
        correlation indices.
    parameters : :class:`biosteam.Parameter`
        Metric(s) of interest for the plot, will be default to all parameters
        included in ``corr_dct`` if not provided.
    metrics : :class:`biosteam.Metric`
        Metric(s) of interest for the plot, will be default to all metrics
        included in ``corr_dct`` if not provided.
    top : int
        Plot the top X parameters with the highest absolute correlation indices,
        this is only applicable for the case of just one metric.
    file : str
        If provided, the generated figure will be saved as a png file.
    close_fig : bool
        Whether to close the figure
        (if not close, new figure will be overlaid on the current figure).
    kwargs
        Other kwargs that will be passed to :func:`seaborn.relplot`.

    Returns
    -------
    figure : :class:`matplotlib.figure.Figure`
        The generated figure.
    axis : :class:`matplotlib.axes._subplots.AxesSubplot`
        The generated figure axis.

    Examples
    --------
    `QSDsan.stats <https://qsdsan.readthedocs.io/en/latest/stats.html>`_

    See Also
    --------
    :func:`biosteam.plots.plot_spearman`

    :func:`seaborn.relplot` and `scatter heat map <https://seaborn.pydata.org/examples/heat_scatter.html>`_
    '''

    df = _update_df_names(result_df)

    param_names = _update_input(parameters, df.index)
    param_names = param_names if isinstance(param_names[0], str) \
                              else [p.name for p in param_names]

    metric_names = _update_input(metrics, df.columns)
    metric_names = metric_names if isinstance(metric_names[0], str) \
                                else [m.name for m in metric_names]

    df = df[metric_names].loc[param_names]

    if len(param_names)*len(metric_names) == 0:
        raise ValueError('No correlation data for plotting.')

    elif len(metric_names) == 1: # one metric, tornado plot
        fig, ax = _plot_corr_tornado(df, top)
        return _save_fig_return(fig, ax, file, close_fig)

    else: # multiple metrics, bubble plot
        corr_df = pd.DataFrame()

        for m in df.columns:
            temp_df = pd.DataFrame(columns=('parameter', 'metric', 'correlation', 'size'))
            temp_df['parameter'] = df.index
            temp_df['metric'] = [m]*temp_df.shape[0]
            temp_df['correlation'] = df[m].values
            temp_df['size'] = np.abs(df[m].values)
            corr_df = pd.concat((corr_df, temp_df))

        g = _plot_corr_bubble(corr_df, len(metric_names)/len(param_names), **kwargs)

        return _save_fig_return(g.fig, g.ax, file, close_fig)



# =============================================================================
# Plot Morris analysis results
# =============================================================================

def plot_morris_results(morris_dct, metric, kind='scatter',
                        x_axis='mu_star',
                        k1=0.1, k2=0.5, k3=1, label_kind='number',
                        file='', close_fig=True, **kwargs):
    '''
    Visualize the results from Morris One-at-A-Time analysis as either scatter
    or bar plots.
    In scatter plots, the x values are :math:`{\mu^*}` and the y values are :math:`{\sigma}`.
    In bar plots, bar length indicate the :math:`{\mu^*}` values with error bars
    representing confidence intervals of the analysis.

    Parameters
    ----------
    morris_dct : dict
        Results dict generated by :func:`morris_analysis`.
    metric : :class:`biosteam.Metric`
        The metric of interest for the plot.
    kind : str
        Either "scatter" (:math:`{\sigma}`) vs. :math:`{\mu^*}` or "bar" (:math:`{\mu^*}` with confidence interval) plot.
    x_axis : str
        X-axis parameter, should be either "mu_star" (the commonly used one) or "mu".
    k1 : float
        The slope to differentiate monotonic (above the line)
        and linear (below the line).
    k2 : float
        The slope to differentiate almost monotonic (above the line)
        and monotonic (below the line).
    k3 : float
        The slope to differentiate non-linear and/or non-monotonic (above the line)
        and almost monotonic (below the line).
    label_kind : str
        How to label the points, can be either "number" (use index number of the result table)
        of "name" (use index name of the result table).
    file : str
        If provided, the generated figure will be saved as a png file.
    close_fig : bool
        Whether to close the figure
        (if not close, new figure will be overlaid on the current figure).
    kwargs
        Other kwargs that will be passed to :func:`morris.horizontal_bar_plot` in ``SALib.plotting``.

    Returns
    -------
    figure : :class:`matplotlib.figure.Figure`
        The generated figure.
    axis : :class:`matplotlib.axes._subplots.AxesSubplot`
        The generated figure axis.

    Examples
    --------
    `QSDsan.stats <https://qsdsan.readthedocs.io/en/latest/stats.html>`_
    '''

    df = morris_dct[metric.name]
    x_data = getattr(df, x_axis)
    y_data = df.sigma
    num = len(x_data)
    if label_kind == 'number':
        labels = range(num)
    elif label_kind == 'name':
        labels = df.index.values
    else:
        raise ValueError(f'label_kind can only be "number" or "name", not "{label_kind}".')

    ax = plt.subplot()
    sns.set_theme(style='ticks')

    if kind == 'scatter':
        ax.scatter(x_data, y_data, color='k')
        for x, y, label in zip(x_data, y_data, labels):
            ax.annotate(label, (x, y), xytext=(10, 10), textcoords='offset points',
                        ha='center')
        x_range = np.arange(-1, np.ceil(ax.get_xlim()[1])+1)
        line1, = ax.plot(x_range, k1*x_range, color='black', linestyle='-.')
        line2, = ax.plot(x_range, k2*x_range, color='black', linestyle='--')
        line3, = ax.plot(x_range, k3*x_range, color='black', linestyle='-')
        ax.legend((line3, line2, line1), (r'$\sigma/\mu^*$'+f'={k3}',
                                          r'$\sigma/\mu^*$'+f'={k2}',
                                          r'$\sigma/\mu^*$'+f'={k1}'),
                  loc='best')
        if x_axis == 'mu_star':
            ax.set_xlim(0,)
        ax.set_ylim(0,)
        ax.set_xlabel(r'$\mu^*$')
        ax.set_ylabel(r'$\sigma$')
        fig = ax.figure

    elif kind == 'bar':
        if x_axis == 'mu':
            raise ValueError('Bar plot can only be made for mu_star, not mu.')
        df = morris_dct[metric.name]
        df['names'] = df.index
        fig = sa_plt_morris.horizontal_bar_plot(ax, df, opts=kwargs)

    # for ax in fig.axes:
    for key in ax.spines.keys():
        ax.spines[key].set(color='k', linewidth=0.5, visible=True)
        ax.grid(False)

    return _save_fig_return(fig, ax, file, close_fig)


def plot_morris_convergence(result_dct, metric, parameters=(), plot_rank=False,
                            kind='line', show_error=True, palette='pastel', file='', close_fig=True):
    '''
    Plot the evolution of :math:`{\mu^*}` or its rank with the number of trajectories.

    Parameters
    ----------
    result_dct : dict
        Result dictionary generated from :func:`qsdsan.stats.morris_till_convergence`
    metric : :class:`biosteam.Metric`
        The metric of interest for the plot.
    parameters : :class:`biosteam.Parameter`
        Single or a iterable of model parameters whose :math:`{\mu^*}` will be
        included in the plot.
        Will be set to all parameters in retult_dct will be used if not provided.
    plot_rank : bool
        If True, will plot rank of :math:`{\mu^*}` instead of value.

        .. note::
            If plot_rank is True, error bars or bands will not be included.
    kind : str
        Can be either 'line' or 'scatter'.
    show_error : bool
        Whether to include the confidence interval in the plot,
        will be bars for scatter plot and bands for line plot.
    palette : string, list, dict, or :class:`matplotlib.colors.Colormap`
        Will be passed on to :func:`seaborn.color_palette`.
    file : str
        If provided, the generated figure will be saved as a png file.
    close_fig : bool
        Whether to close the figure
        (if not close, new figure will be overlaid on the current figure).

    Returns
    -------
    figure : :class:`matplotlib.figure.Figure`
        The generated figure.
    axis : :class:`matplotlib.axes._subplots.AxesSubplot`
        The generated figure axis.

    Examples
    --------
    `QSDsan.stats <https://qsdsan.readthedocs.io/en/latest/stats.html>`_
    '''

    ax = plt.subplot()
    df = result_dct['mu_star'][metric.name].copy().astype('float64')
    conf_df = result_dct['mu_star_conf'][metric.name].copy().astype('float64')

    param_names = _update_input(parameters, df.columns)
    param_names = param_names if isinstance(param_names[0], str) \
                              else [p.name for p in param_names]

    if plot_rank:
        df = df.rank(axis=1)
        ylabel = f'Rank for {metric.name.lower()}'
        loc = 'lower left'
    else:
        ylabel = f'$\mu^*$ for {metric.name.lower()}'
        loc = 'best'

    palette = sns.color_palette('deep', n_colors=len(param_names))
    sns.set_theme(style='ticks', palette=palette)

    for n, param in enumerate(param_names):
        if kind == 'line':
            ax.plot(df.index, df[param], color=palette[n], linewidth=1.5, label=param)
            if not plot_rank and show_error:
                ax.fill_between(df.index, df[param]-conf_df[param], df[param]+conf_df[param],
                                color=palette[n], linewidth=0, alpha=0.2)
        elif kind == 'scatter':
            ax.scatter(df.index, df[param], color=palette[n], label=param)
            if not plot_rank and show_error:
                ax.errorbar(df.index, df[param], conf_df[param],
                            color=palette[n], alpha=0.5)
        else:
            raise ValueError(f'kind can only be "line" or "scatter", not "{kind}".')

    ax.legend(loc=loc)
    ax.set(xlabel='Number of trajectories', ylabel=ylabel, ylim=(0, ax.get_ylim()[1]))

    return _save_fig_return(ax.figure, ax, file, close_fig)


# =============================================================================
# Plot variance breakdown
# =============================================================================

def _plot_bar(kind, df, error, ax=None):
    ax = ax if ax else plt.subplot()

    sns.set_theme(style='white')
    if 'ST' in kind:
        sns.set_color_codes('pastel')
        sns.barplot(x=df.ST, y=df.index, data=df,
                    ax=ax, label='Total', color='b')
        if error:
            ax.errorbar(df.ST, df.index, xerr=df.ST_conf, fmt='none', ecolor='b')

    if 'S1' in kind:
        sns.set_color_codes('muted')
        sns.barplot(x=df.S1, y=df.index, data=df,
                    ax=ax, label='Main', color='b')
        if error:
            ax.errorbar(df.S1, df.index, xerr=df.S1_conf, fmt='none', ecolor='b')

    ax.set_xlabel('Variance')
    ax.legend(ncol=2, loc='lower right', frameon=True)
    ax.set_ylim(df.shape[0]-0.5, -0.5)

    return ax

def _plot_heatmap(hmap_df, ax=None, annot=False, diagonal='', sts1_df=None,
                  default_cbar=True):
    ax = ax if ax else plt.subplot()
    ax_cbar = ax.figure.add_axes([0.03, 0.3, 0.02, 0.4]) if not default_cbar else None

    if diagonal:
        np.fill_diagonal(hmap_df.values, getattr(sts1_df, diagonal))
        hmap_df = hmap_df.astype('float64')
        k = -1
        title = 'Total/Interaction Effects' if diagonal=='ST' else 'Main/Interaction Effects'
    else:
        hmap_df = hmap_df.fillna(0)
        k = 0
        title = 'Interaction Effects'

    mask = np.tril(np.ones_like(hmap_df, dtype=bool), k)

    sns.set_theme(style='white')
    cmap = sns.diverging_palette(230, 20, as_cmap=True)
    sns.heatmap(hmap_df,
                mask=mask,
                ax=ax, cmap=cmap, center=0, linewidths=.5,
                annot=annot, cbar_kws={'shrink': 0.5}, cbar_ax=ax_cbar)
    ax.set_title(title)

    return ax


def plot_fast_results(result_dct, metric, parameters=(), error_bar=True,
                      file='', close_fig=True):
    '''
    Visualize the results from FAST or RBD-FAST analysis as a bar plot.

    Parameters
    ----------
    result_dct : dict
        Result dictionary generated from :func:`qsdsan.stats.fast_analysis`
    metric : :class:`biosteam.Metric`
        The metric of interest for the plot.
    parameters : :class:`biosteam.Parameter`
        Single or a iterable of model parameters whose :math:`{\mu^*}` will be
        included in the plot.
        Will be set to all parameters in retult_dct will be used if not provided.
    error_bar : bool
        Whether to include the confidence interval as error bars in the plot.
    file : str
        If provided, the generated figure will be saved as a png file.
    close_fig : bool
        Whether to close the figure
        (if not close, new figure will be overlaid on the current figure).

    Returns
    -------
    figure : :class:`matplotlib.figure.Figure`
        The generated figure.
    axis : :class:`matplotlib.axes._subplots.AxesSubplot`
        The generated figure axis.

    Examples
    --------
    `QSDsan.stats <https://qsdsan.readthedocs.io/en/latest/stats.html>`_
    '''

    param_names = _update_input(parameters, result_dct[metric.name]['S1'].index)
    param_names = param_names if isinstance(param_names[0], str) \
                              else [p.name for p in param_names]

    df = result_dct[metric.name]
    kind = 'STS1' if 'ST' in df.columns else 'S1'

    ax = _plot_bar(kind, df, error_bar)

    return _save_fig_return(ax.figure, ax, file, close_fig)


def plot_sobol_results(result_dct, metric, parameters=(), kind='all',
                       annotate_heatmap=False, plot_in_diagonal='',
                       error_bar=True, file='', close_fig=True):
    '''
    Visualize the results from Sobol analysis as a bar plot and/or heat map.
    Total (:math:`S_{Ti}`) and main (:math:`S_{1i}`) effects can be drawn in
    the bar plot or diagnoal of the heat map;
    second-order interaction effects between two parameters (:math:`S_{2ij}`)
    are shown in the heat map.

    Parameters
    ----------
    result_dct : dict
        Result dictionary generated from :func:`qsdsan.stats.sobol_analysis`
    metric : :class:`biosteam.Metric`
        The metric of interest for the plot.
    parameters : :class:`biosteam.Parameter`
        Single or a iterable of model parameters whose :math:`{\mu^*}` will be
        included in the plot.
        Will be set to all parameters in retult_dct will be used if not provided.
    kind : str
        Which sensitivity index or indices to plot:

        +------+---------------------------------------------------------------+
        | kind | returned plot                                                 |
        +======+===============================================================+
        | ST   | total effects (bar)                                           |
        +------+---------------------------------------------------------------+
        | S1   | main effects (bar)                                            |
        +------+---------------------------------------------------------------+
        | S2   | interaction effects (heat map)                                |
        +------+---------------------------------------------------------------+
        | STS1 | total and main effects (bar)                                  |
        +------+---------------------------------------------------------------+
        | STS2 | total and interaction effects (heat map or bar and heat map)  |
        +------+---------------------------------------------------------------+
        | S1S2 | main and interaction effects (heat map or bar and heat map)   |
        +------+---------------------------------------------------------------+
        | all  | all effects (bar and heat map)                                |
        +------+---------------------------------------------------------------+
    annotate_heatmap : bool
        Whether to annotate the index values in the heat map.
    plot_in_diagonal : str
        Plot total or main effects in the diagonal of the interaction heat map,
        can be "ST", "S1", or "".
        This is applicable when kind is "STS2", "S1S2", or "all".
    error_bar : bool
        Whether to include the confidence interval as error bars in the plot,
        this is only applicable for the bar plot.
    file : str
        If provided, the generated figure will be saved as a png file.
    close_fig : bool
        Whether to close the figure
        (if not close, new figure will be overlaid on the current figure).

    Returns
    -------
    figure : :class:`matplotlib.figure.Figure`
        The generated figure.
    axis : :class:`matplotlib.axes._subplots.AxesSubplot`
        The generated figure axis.
        If generating bar plot and heat map, a tuple of two axes will be returned
        for the respective plot.

    Examples
    --------
    `QSDsan.stats <https://qsdsan.readthedocs.io/en/latest/stats.html>`_
    '''

    kind_upper = kind.upper()
    if kind_upper=='ALL' or set(kind_upper)==set('STS1S2'):
        kind_upper = 'STS1S2'
    elif not set(kind_upper).union(set('STS1S2'))==set('STS1S2'):
        raise ValueError(f'The plot kind of "{kind}" is invalid.')

    ax_sts1 = ax_s2 = None

    param_names = _update_input(parameters, result_dct[metric.name]['ST'].index)
    param_names = param_names if isinstance(param_names[0], str) \
                              else [p.name for p in param_names]

    st_df = result_dct[metric.name]['ST'].loc[[p for p in param_names]]
    s1_df = result_dct[metric.name]['S1'].loc[[p for p in param_names]]
    sts1_df = pd.concat((st_df, s1_df), axis=1).sort_values('ST', ascending=False)

    if kind_upper in ('ST', 'S1', 'STS1', 'S1ST'): # no S2, bar plot only
        ax_sts1 = _plot_bar(kind_upper, sts1_df, error_bar)
        return _save_fig_return(ax_sts1.figure, ax_sts1, file, close_fig)

    else: # has S2, need heat map
        s2_df = result_dct[metric.name]['S2']
        hmap_df = pd.DataFrame(columns=sts1_df.index, index=sts1_df.index)
        for (p1, p2) in s2_df.index:
            if not (p1 in hmap_df.index and p2 in hmap_df.index):
                continue
            hmap_df[p1][p2] = hmap_df[p2][p1] = s2_df.S2[(p1, p2)]

        if kind_upper == 'S2': # only S2, only heat map
            ax_s2 = _plot_heatmap(hmap_df, annot=annotate_heatmap)
            return _save_fig_return(ax_s2.figure, ax_s2, file, close_fig)

        else: # has S2, need heat map
            not_s2 = ''.join(i for i in kind_upper.split('S2')) # 'ST', 'S1', or 'STS1'

            plot_in_diagonal = plot_in_diagonal.upper()
            if not_s2 and not_s2 == plot_in_diagonal: # ST or S1 in heat map, only heat map
                ax_s2 = _plot_heatmap(hmap_df, ax=ax_s2, annot=annotate_heatmap,
                                      diagonal=plot_in_diagonal,
                                      sts1_df=sts1_df, default_cbar=True)
                return _save_fig_return(ax_s2.figure, ax_s2, file, close_fig)

            # ST or S1 in bar, need bar and heat map
            elif not_s2 != 'STS1':
                warn(f'The plot_in_diagonal value of "{plot_in_diagonal}" ' \
                     f'is invalid for kind "{kind}" and is ignored.')
                plot_in_diagonal = ''

            if plot_in_diagonal and plot_in_diagonal.upper() not in ('ST', 'S1'):
                raise ValueError('plot_in_diagonal must be "ST", "S1", or "", '\
                                 f'not "{plot_in_diagonal}".')

            fig, (ax_s2, ax_sts1) = plt.subplots(1, 2, figsize=(8, 5))
            bar = not_s2.replace(plot_in_diagonal, '') # 'ST', 'S1', or 'STS1'

            ax_sts1 = _plot_bar(bar, sts1_df, error_bar, ax=ax_sts1)
            ax_sts1.yaxis.set_visible(False)
            if bar == 'ST':
                ax_sts1_title = 'Total Effects'
            elif bar == 'S1':
                ax_sts1_title = 'Main Effects'
            else:
                ax_sts1_title = 'Total/Main Effects'
            ax_sts1.set_title(ax_sts1_title)

            ax_s2 = _plot_heatmap(hmap_df, ax=ax_s2, annot=annotate_heatmap,
                                        diagonal=plot_in_diagonal,
                                        sts1_df=sts1_df, default_cbar=False)

            labels = [i if len(i)<=15 else i[0:15]+'...' for i in hmap_df.index]
            ax_s2.yaxis.set_label_position('right')
            ax_s2.yaxis.tick_right()
            ax_s2.set_yticklabels(labels, rotation=0, ha='center',
                                  position=(1.2, 0.0))

            ax_s2.tick_params(length=0)
            xlabels = labels.copy()
            xlabels[0] = '' if not plot_in_diagonal else xlabels[0]
            ax_s2.set_xticklabels(xlabels)

            plt.tight_layout()
            plt.subplots_adjust(wspace=0.4, top=0.85)
            fig.suptitle(f'Variance breakdown for {metric.name.lower()}')

        return _save_fig_return(fig, (ax_sts1, ax_s2), file, close_fig)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''

# %%

from collections.abc import Iterable
from biosteam._graphics import splitter_graphics
from .. import SanUnit

__all__ = ('ComponentSplitter',)


class ComponentSplitter(SanUnit):
    '''
    Split the influent into individual components,
    the last effluent contains all remaining components.

    Parameters
    ----------
    split_keys : iterable
        IDs of components to be splitted to different effluents.
        Element of the item in the iterable can be str or another iterable
        containing component IDs.
        If the item is also iterable, all components whose ID are in the iterable
        will be splitted to the same effluent.
        The split is always 1 for a certain component to an effluent (i.e., complete split).

        .. note::

            Length of the `split_keys()` (which determines size of the outs) \
            cannot be changed after initiation.

    Examples
    --------
    `bwaise systems <https://github.com/QSD-Group/EXPOsan/blob/main/exposan/bwaise/systems.py>`_
    '''

    def __init__(self, ID='', ins=None, outs=(), thermo=None,
                 init_with='WasteStream', split_keys=()):
        if not split_keys:
            raise ValueError('`split_keys` cannot be empty.')

        if isinstance(split_keys, str):
            self._N_outs = 2
        else:
            self._N_outs = len(split_keys) + 1
        SanUnit.__init__(self, ID, ins, outs, thermo, init_with)

        self._split_keys = split_keys


    _ins_size_is_fixed = False
    _outs_size_is_fixed = False
    _graphics = splitter_graphics


    def _run(self):
        last = self.outs[-1]
        last.mix_from(self.ins)

        splitted = []
        for num, cmps in enumerate(self.split_keys):
            if isinstance(cmps, str):
                cmps = (cmps,)

            elif not isinstance(cmps, Iterable):
                raise ValueError('`split_keys` must be an iterable, '
                                 f'not {type(cmps).__name__}.')

            for cmp in cmps:
                self.outs[num].imass[cmp] = last.imass[cmp]
                last.imass[cmp] = 0
                if cmp in splitted:
                    raise ValueError(f'The component {cmps} appears more than once in `split_keys`.')
                splitted.append(cmp)


    @property
    def split_keys(self):
        '''
        [iterable] IDs of components to be splitted to different effluents.
        Element of the item in the iterable can be str or another iterable
        containing component IDs.
        If the item is also iterable, all components whose ID are in the iterable
        will be splitted to the same effluent.
        The split is always 1 for a certain component to an effluent (i.e., complete split).

        .. note::

            Length of the `split_keys()` (which determines size of the outs) \
                cannot be changed after initiation.
        '''
        return self._split_keys
    @split_keys.setter
    def split_keys(self, i):
        if isinstance(i, str):
            i = (i,)

        if len(i) != len(self.outs):
            raise ValueError('Size of `split_keys` cannot be changed after initiation.')

        self._split_keys = i


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


# %%

from warnings import warn
from .. import SanUnit
from ._decay import Decay
from ..utils import load_data, data_path, dct_from_str

__all__ = ('SludgeSeparator',)

data_path += 'sanunit_data/_sludge_separator.tsv'

allocate_N_removal = Decay.allocate_N_removal

class SludgeSeparator(SanUnit):
    '''
    For sludge separation based on Trimmer et al. [1]_, note that no default
    cost or environmental impacts are included.

    Parameters
    ----------
    ins : WasteStream
        Waste for treatment.
    outs : WasteStream
        Liquid, settled solids.
    split : float or dict
        Fractions of material retention in the settled solids.
        Default values will be used if not given.
    settled_frac : float
        Fraction of influent that settles as solids.
        The default value will be used if not given.

    Examples
    --------
    `bwaise systems <https://github.com/QSD-Group/EXPOsan/blob/main/exposan/bwaise/systems.py>`_

    References
    ----------
    .. [1] Trimmer et al., Navigating Multidimensional Social–Ecological System
        Trade-Offs across Sanitation Alternatives in an Urban Informal Settlement.
        Environ. Sci. Technol. 2020, 54 (19), 12641–12653.
        https://doi.org/10.1021/acs.est.0c03296.

    '''

    def __init__(self, ID='', ins=None, outs=(), thermo=None, init_with='WasteStream',
                 split=None, settled_frac=None, **kwargs):
        SanUnit.__init__(self, ID, ins, outs, thermo, init_with, **kwargs)

        data = load_data(path=data_path)
        if not split:
            value = dct_from_str(data.loc['split']['expected'])
            setattr(self, 'split', value)
        if not settled_frac:
            value = float(data.loc['settled_frac']['expected'])
            setattr(self, 'settled_frac', value)
        del data

    _N_ins = 1
    _outs_size_is_fixed = False

    def _adjust_solid_water(self, influent, liq, sol):
        sol.imass['H2O'] = 0
        sol.imass['H2O'] = influent.F_mass * self.settled_frac - sol.F_mass
        if sol.imass['H2O'] < 0:
            sol.imass['H2O'] = 0
            msg = 'Negative water content calcualted for settled solids, ' \
                'try smaller split or larger settled_frac.'
            warn(msg, stacklevel=2)
        liq.imass['H2O'] = influent.imass['H2O'] - sol.imass['H2O']
        return liq, sol

    def _run(self):
        waste = self.ins[0]
        liq, sol = self.outs[0], self.outs[1]

        # Retention in the settled solids
        split = self.split
        if self._split_type == 'float':
            liq.copy_like(waste)
            sol.copy_like(waste)
            sol.mass *= self.split
            liq.mass -= sol.mass
        else:
            for var in self.split.keys():
                #!!! In the future this should be best done by changing the state variable
                if var == 'TS':
                    try: sol.imass['OtherSS'] = split[var] * waste.imass['OtherSS']
                    except: breakpoint()
                elif var == 'COD':
                    sol_COD = split[var] * waste._COD * waste.F_vol
                    liq_COD = waste._COD * waste.F_vol - sol_COD
                elif var == 'N':
                    N_sol = split[var]*(waste.imass['NH3']+waste.imass['NonNH3'])
                    NonNH3_rmd, NH3_rmd = \
                        allocate_N_removal(N_sol, waste.imass['NonNH3'])
                    sol.imass['NonNH3'] = NonNH3_rmd
                    sol.imass['NH3'] = NH3_rmd
                else:
                    sol.imass[var] = split[var] * waste.imass[var]
            liq.mass = waste.mass - sol.mass

        # Adjust total mass of of the settled solids by changing water content.
        liq, sol = self._adjust_solid_water(waste, liq, sol)
        sol._COD = sol_COD / sol.F_vol
        liq._COD = liq_COD / liq.F_vol


    @property
    def split(self):
        '''
        [float] or [dict] Fractions of material retention in the settled solids
        before degradation. If a single number is provided, then it is assumed
        that retentions of all Components in the WasteStream are the same.

        .. note::

            Set state variable values (e.g., COD) will be retained if the retention
            ratio is a single number (treated like the loss stream is split
            from the original stream), but not when the ratio is a dict.

        '''
        return self._split
    @split.setter
    def split(self, i):
        try:
            self._split = float(i)
            self._split_type = 'float'
        except:
            if isinstance(i, dict):
                self._split = i
                self._split_type = 'dict'
            else:
                raise TypeError(f'Only float or dict allowed, not {type(i).__name__}.')

    @property
    def settled_frac(self):
        '''[float] Fraction of influent that settles as solids.'''
        return self._settled_frac
    @settled_frac.setter
    def settled_frac(self, i):
        self._settled_frac = i


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


# %%

__all__ = ('FuncGetter',)


# # The below two classes haven't been successfully used,
# # not sure if the idea is good.
# class AttrGetter:
#     __slots__ = ('attr',)
#     def __init__(self, obj, attr):
#         self.attr = getattr(obj, attr)

#     def __call__(self):
#         return self.attr


# class AttrGetter:
#     __slots__ = ('obj', 'attrs')
#     def __init__(self, obj, attrs):
#         self.obj = obj
#         if isinstance(attrs, str):
#             attrs = (attrs,)
#         self.attrs = attrs

#     def __call__(self):
#         if len(self.attrs) == 1:
#             print(getattr(self.obj, self.attrs[0]))
#             return getattr(self.obj, self.attrs[0])
#         return [getattr(self.obj, attr) for attr in self.attrs]


class FuncGetter:
    __slots__ = ('func', 'params')
    def __init__(self, func, params):
        self.func = func
        self.params = params

    def __call__(self):
        return self.func(*self.params)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''


# %%

from warnings import warn
from .. import SanUnit

__all__ = ('CropApplication',)

class CropApplication(SanUnit):
    '''
    Recovery nutrients in the recycled excreta (energy not recovered) based on Trimmer et al. [1]_

    Parameters
    ----------
    if_material_loss : bool or dict
        If material loss occurs during application.
    loss_ratio : float or dict
        Fractions of material losses during application (if `if_materiloass` is True).

    Examples
    --------
    `bwaise systems <https://github.com/QSD-Group/EXPOsan/blob/main/exposan/bwaise/systems.py>`_

    References
    ----------
    .. [1] Trimmer et al., Navigating Multidimensional Social–Ecological System
        Trade-Offs across Sanitation Alternatives in an Urban Informal Settlement.
        Environ. Sci. Technol. 2020, 54 (19), 12641–12653.
        https://doi.org/10.1021/acs.est.0c03296.
    '''

    def __init__(self, ID='', ins=None, outs=(), thermo=None, init_with='WasteStream',
                 if_material_loss=True, loss_ratio=0.02):

        SanUnit.__init__(self, ID, ins, outs, thermo, init_with)
        self.if_material_loss = if_material_loss
        self.loss_ratio = loss_ratio

    _N_ins = 1
    _N_outs = 2

    def _run(self):
        applied, loss = self.outs
        applied.copy_like(self.ins[0])
        loss.empty()
        if self.if_material_loss:
            if self._loss_ratio_type == 'float':
                loss.copy_like(applied)
                applied.mass *= 1 - self.loss_ratio
                loss.mass = self.ins[0].mass - applied.mass
            else:
                for cmp, ratio in self.loss_ratio.items():
                    applied.imass[cmp] *= 1 - ratio
                    loss.imass[cmp] = self.ins[0].imass[cmp] - applied.imass[cmp]


    @property
    def loss_ratio(self):
        '''
        [float] or [dict] Fractions of material losses during application.
        If a single number is provided, then it is assumed that losses of
        all Components in the WasteStream are the same.

        .. note::

            Set state variable values (e.g., COD) will be retained if the loss
            ratio is a single number (treated like the loss stream is split
            from the original stream), but not when the ratio is a dict.

        '''
        return self._loss_ratio
    @loss_ratio.setter
    def loss_ratio(self, i):
        if not self.if_material_loss:
            msg = f'`if_material_loss` is False, the set value {i} is ignored.'
            warn(msg, source=self)
        else:
            try:
                self._loss_ratio = float(i)
                self._loss_ratio_type = 'float'
            except TypeError:
                if isinstance(i, dict):
                    self._loss_ratio = i
                    self._loss_ratio_type = 'dict'
                else:
                    raise TypeError(f'Only float or dict allowed, not {type(i).__name__}.')


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Yalin Li <zoe.yalin.li@gmail.com>
    Joy Cheung <joycheung1994@gmail.com>

Part of this module is based on the Thermosteam package:
https://github.com/BioSTEAMDevelopmentGroup/thermosteam

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''

import numpy as np
import pandas as pd
import thermosteam as tmo
from . import _component, Chemical, Chemicals, CompiledChemicals, Component
from .utils import load_data

__all__ = ('Components', 'CompiledComponents')

setattr = object.__setattr__

utils = tmo.utils
_component_properties = _component._component_properties
_num_component_properties = _component._num_component_properties
_key_component_properties = _component._key_component_properties
# _TMH = tmo.base.thermo_model_handle.ThermoModelHandle
_PH = tmo.base.phase_handle.PhaseHandle


# %%

class UndefinedComponent(AttributeError):
    '''AttributeError regarding undefined :class:`Component` objects.'''
    def __init__(self, ID):
        super().__init__(repr(ID))

def must_compile(*args, **kwargs): # pragma: no cover
    raise TypeError('Method valid only for CompiledChemicals, '
                    'run <Components>.compile() to compile first.')


# =============================================================================
# Define the Components class
# =============================================================================

class Components(Chemicals):
    '''
    A subclass of :class:`thermosteam.Chemicals`, contains :class:`Component`
    objects as attributes.

    Examples
    --------
    `Component and WasteStream <https://qsdsan.readthedocs.io/en/latest/tutorials/Component_and_WasteStream.html>`_

    See Also
    --------
    `thermosteam.Chemicals <https://thermosteam.readthedocs.io/en/latest/Chemicals.html>`_

    '''

    def __new__(cls, components, cache=False):
        self = super(Chemicals, cls).__new__(cls)
        isa = isinstance
        setfield = setattr
        CASs = set()
        for i in components:
            if isa(i, Component):
                CAS = i.CAS
                if CAS in CASs: continue
                CASs.add(CAS)
                setfield(self, i.ID, i)
            elif isa(i, Chemical):
                raise TypeError(f'{i} is a `thermosteam.Chemical` object, '
                                'use `Component.from_chemical` to define a `Component` object.')
            else:
                raise TypeError(f'Only `Component` objects can be included, not a `{type(i).__name__}` object.')

        return self


    def __setattr__(self, ID, component):
        raise TypeError('Cannot set attribute; use `<Components>.append/extend` instead.')


    def __setitem__(self, ID, component):
        raise TypeError('Cannot set item; use `<Components>.append/extend` instead.')


    def __getitem__(self, key):
        '''Return a :class:`Component` object or a list of :class:`Component` objects.'''
        dct = self.__dict__
        try:
            if isinstance(key, str):
                return dct[key]
            else:
                return [dct[i] for i in key]
        except KeyError as key_error:
            raise UndefinedComponent(key_error.args[0])


    def __contains__(self, component):
        if isinstance(component, str):
            return component in self.__dict__
        elif isinstance(component, Component):
            return component in self.__dict__.values()
        else: # pragma: no cover
            return False


    def __repr__(self):
        return f"Components([{', '.join(self.__dict__)}])"


    def copy(self):
        '''Return a copy.'''
        copy = object.__new__(Components)
        for cmp in self: setattr(copy, cmp.ID, cmp)
        return copy


    def append(self, component):
        '''Append a Component'''
        if not isinstance(component, Component):
            if isinstance(component, Chemical):
                raise TypeError(f'{component} is a `Chemical` object, '
                                'use `Component.from_chemical` to define a `Component` object.')
            else:
                raise TypeError("only `Component` objects can be appended, "
                               f"not `{type(component).__name__}` object.")
        ID = component.ID
        if ID in self.__dict__:
            raise ValueError(f"{ID} already defined in this `Components` object.")
        setattr(self, ID, component)


    def extend(self, components):
        '''Extend with more :class:`Component` objects.'''
        if isinstance(components, Components):
            self.__dict__.update(components.__dict__)
        else:
            for component in components: self.append(component)


    def compile(self, skip_checks=False):
        '''Cast as a :class:`CompiledComponents` object.'''
        components = tuple(self)
        setattr(self, '__class__', CompiledComponents)
        try: self._compile(components, skip_checks)
        except Exception as error:
            setattr(self, '__class__', Chemicals)
            setattr(self, '__dict__', {i.ID: i for i in components})
            raise error

    kwarray = array = index = indices = must_compile

    _default_data = None


    @classmethod
    def load_from_file(cls, path_or_df, index_col=None,
                       use_default_data=False, store_data=False):
        '''
        Create and return a :class:`Components` objects based on properties
        defined in a datasheet.

        Parameters
        ----------
        path_or_df : str or :class:`pandas.DataFrame`
            File path, the file should end with ".cvs", ".xls", or "xlsx".
        index_col : None or int
            Index column of the :class:`pandas.DataFrame`.
        use_default_data : bool
            Whether to use the default components.
        store_data : bool
            Whether to store this as the default components.

        Returns
        -------
        A :class:`Components` object that contains all created Component objects.


        .. note::

            The :class:`Components` object needs to be compiled before it is used in simulation.

        '''
        if use_default_data and cls._default_data is not None:
            data = cls._default_data
        elif isinstance(path_or_df, str):
            data = load_data(path_or_df, index_col=index_col)
        else:
            data = path_or_df

        new = cls(())

        for i, cmp in data.iterrows():
            if pd.isna(cmp.measured_as):
                cmp.measured_as = None
            try:
                component = Component(ID = cmp.ID,
                                      search_ID = str(cmp.CAS),
                                      measured_as = cmp.measured_as)
            except LookupError:
                try:
                    component = Component(ID = cmp.ID,
                                          search_ID = 'PubChem='+str(int(cmp.PubChem)),
                                          measured_as = cmp.measured_as)
                except:
                    if not pd.isna(cmp.formula):
                        component = Component(ID = cmp.ID,
                                              formula = cmp.formula,
                                              measured_as = cmp.measured_as)
                    else:
                        component = Component(ID = cmp.ID,
                                              measured_as = cmp.measured_as)
            for j in _component_properties:
                field = '_' + j
                if pd.isna(cmp[j]): setattr(component, j, None)
                else: setattr(component, field, cmp[j])
            new.append(component)

        if store_data:
            cls._default_data = data
        return new


    @classmethod
    def load_default(cls, use_default_data=True, store_data=True, default_compile=True):
        '''
        Create and return a :class:`Components` or :class:`CompiledComponents`
        object containing all default :class:`Component` objects.

        Parameters
        ----------
        use_default_data : bool, optional
            Whether to use default cache data. The default is True.
        store_data : bool, optional
            Whether to store the default data as cache. The default is True.
        default_compile : bool, optional
            Whether to compile the default :class:`Components`. The default is True.

        Returns
        -------
        A :class:`Components` or :class:`CompiledComponents` object with
        default :class:`Component` objects.


        .. note::

            [1] Component-specific properties are defined in ./data/component.cvs.

            [2] When `default_compile` is True, all essential chemical-specific properties
            (except molar volume model and normal boiling temperature) that are missing will
            be defaulted to those of water.

            [3] When `default_compile` is True, missing molar volume models will be defaulted
            according to particle sizes: particulate or colloidal -> 1.2e-5 m3/mol,
            soluble -> copy from urea, dissolved gas -> copy from CO2.

            [4] When `default_compile` is True, missing normal boiling temoerature will be
            defaulted according to particle sizes: particulate or colloidal -> copy from NaCl,
            soluble -> copy from urea, dissolved gas -> copy from CO2.


        '''
        import os
        path = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data/_components.tsv')
        del os
        new = cls.load_from_file(path, index_col=None, use_default_data=True, store_data=True)

        H2O = Component.from_chemical('H2O', Chemical('H2O'),
                                      i_charge=0, f_BOD5_COD=0, f_uBOD_COD=0,
                                      f_Vmass_Totmass=0, description="Water",
                                      particle_size='Soluble',
                                      degradability='Undegradable', organic=False)
        new.append(H2O)

        if default_compile:
            isa = isinstance
            for i in new:
                i.default()

                if i.particle_size == 'Soluble':
                    ref_chem = Chemical('urea')
                elif i.particle_size == 'Dissolved gas':
                    ref_chem = Chemical('CO2')
                else:
                    ref_chem = Chemical('NaCl')

                i.Tb = ref_chem.Tb if not i.Tb else i.Tb

                COPY_V = False # if don't have V model, set those to default
                if isa(i.V, _PH):
                    if not i.V.l.valid_methods(298.15):
                        COPY_V = True
                else:
                    if not i.V.valid_methods(298.15):
                        COPY_V = True

                if COPY_V:
                    if i.particle_size in ('Soluble', 'Dissolved gas'):
                        i.copy_models_from(ref_chem, names=('V',))
                    else:
                        try: i.V.add_model(1.2e-5)
                        except AttributeError:
                            i.V.l.add_model(1.2e-5)    # m^3/mol
                            i.V.s.add_model(1.2e-5)

                i.copy_models_from(H2O)

                try:
                    i.Hvap(i.Tb)
                except RuntimeError: # Hvap model of H2O cannot be extrapolated to Tb
                    i.copy_models_from(ref_chem, names=('Hvap',))

            new.compile()
        return new


    @classmethod
    def from_chemicals(cls, chemicals, **data):
        '''
        Return a new :class:`Components` from a :class:`thermosteam.Chemicals`
        or :class:`thermosteam.CompiledChemicals` object.

        Parameters
        ----------
        chemicals: thermosteam.Chemicals
            The :class:`thermosteam.Chemicals` object as the basis
            for the new :class:`~.Components` object.
            :class:`Component` objects will have the same ID as the corresponding
            :class:`thermosteam.Chemical` object in the :class:`thermosteam.Chemicals`
            object.
        data : dict
            A nested dict with keys being the new components and values being the inner dict,
            keys and values of the inner dict are the attribute names and values, respectively.

        Examples
        --------
        >>> import qsdsan as qs
        >>> chems = qs.Chemicals((qs.Chemical('Water'), qs.Chemical('Ethanol')))
        >>> data = {'Water': {'particle_size': 'Soluble',
        ...                   'degradability': 'Undegradable',
        ...                   'organic': False},
        ...         'Ethanol': {'particle_size': 'Soluble',
        ...                     'degradability': 'Readily',
        ...                     'organic': False}}
        >>> cmps = qs.Components.from_chemicals(chems, **data)
        >>> cmps
        Components([Water, Ethanol])
        '''
        cmps = cls.__new__(cls, ())
        for i in chemicals.__iter__():
            val_dct = data.get(i.ID)
            cmp = Component.from_chemical(i.ID, i)
            if val_dct:
                for k, v in val_dct.items():
                    setattr(cmp, k, v)
            cmps.append(cmp)

        return cmps


# %%

# =============================================================================
# Define the CompiledComponents class
# =============================================================================

chemical_data_array = tmo._chemicals.chemical_data_array

def component_data_array(components, attr):
    data = chemical_data_array(components, attr)
    return data


class CompiledComponents(CompiledChemicals):
    '''
    A subclass of :class:`thermosteam.CompiledChemicals`, contains `Component` objects as attributes.

    See Also
    --------
    `thermosteam.CompiledChemicals <https://thermosteam.readthedocs.io/en/latest/Chemicals.html>`_

    '''

    _cache = {}

    def __new__(cls, components, cache=None):
        isa = isinstance
        components = tuple([cmp if isa(cmp, Component) else Component(cmp, cache)
                           for cmp in components])
        cache = cls._cache
        if components in cache:
            self = cache[components]
        else:
            self = object.__new__(cls)
            setfield = setattr
            for cmp in components:
                setfield(self, cmp.ID, cmp)
            self._compile()
            cache[components] = self
        return self


    def __contains__(self, component):
        if isinstance(component, str):
            return component in self.__dict__
        elif isinstance(component, Component):
            return component in self.tuple
        else: # pragma: no cover
            return False


    def __repr__(self):
        return f"CompiledComponents([{', '.join(self.IDs)}])"


    def refresh_constants(self):
        '''
        Refresh constant arrays of :class:`Components` objects,
        including all chemical and component-specific properties.
        '''
        super().refresh_constants()
        dct = self.__dict__
        components = self.tuple
        for i in _num_component_properties:
            dct[i] = component_data_array(components, i)

    def compile(self, skip_checks=False):
        '''Skip, :class:`CompiledComponents` have already been compiled.'''
        pass


    def _compile(self, components, skip_checks=False):
        dct = self.__dict__
        tuple_ = tuple # this speeds up the code
        components = tuple_(dct.values())
        CompiledChemicals._compile(self, components, skip_checks)
        for component in components:
            missing_properties = component.get_missing_properties(_key_component_properties)
            if not missing_properties: continue
            missing = utils.repr_listed_values(missing_properties)
            raise RuntimeError(f'{component} is missing key component-related properties ({missing}).')

        for i in _num_component_properties:
            dct[i] = component_data_array(components, i)

        dct['s'] = np.asarray([1 if cmp.particle_size == 'Soluble' else 0 for cmp in components])
        dct['c'] = np.asarray([1 if cmp.particle_size == 'Colloidal' else 0 for cmp in components])
        dct['x'] = np.asarray([1 if cmp.particle_size == 'Particulate' else 0 for cmp in components])
        dct['b'] = np.asarray([1 if cmp.degradability != 'Undegradable' else 0 for cmp in components])
        dct['rb'] = np.asarray([1 if cmp.degradability == 'Readily' else 0 for cmp in components])
        dct['org'] = np.asarray([int(cmp.organic) for cmp in components])


    def subgroup(self, IDs):
        '''Create a new subgroup of :class:`Component` objects.'''
        components = self[IDs]
        new = Components(components)
        new.compile()
        for i in new.IDs:
            for j in self.get_synonyms(i):
                try: new.set_synonym(i, j)
                except: pass
        return new


    def index(self, ID):
        '''Return index of specified component.'''
        try: return self._index[ID]
        except KeyError:
            raise UndefinedComponent(ID)


    def indices(self, IDs):
        '''Return indices of multiple components.'''
        try:
            dct = self._index
            return [dct[i] for i in IDs]
        except KeyError as key_error:
            raise UndefinedComponent(key_error.args[0])


    def copy(self):
        '''Return a copy.'''
        copy = Components(self)
        copy.compile()
        return copy


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''
QSDsan: Quantitative Sustainable Design for sanitation and resource recovery systems

This module is developed by:
    Joy Cheung <joycheung1994@gmail.com>

This module is under the University of Illinois/NCSA Open Source License.
Please refer to https://github.com/QSD-Group/QSDsan/blob/main/LICENSE.txt
for license details.
'''
from chemicals.elements import molecular_weight

__all__ = ('cod_test_stoichiometry', 'electron_acceptor_cod')



#%%

dichromate_oxidizing_elements = ('C', 'H', 'O', 'N', 'S', 'P')
dichromate_oxidizing_elements_set = frozenset(dichromate_oxidizing_elements)

def cod_test_stoichiometry(atoms, charge=0, MW=None, missing_handling='elemental'):
    r'''
    Return a dictionary of stoichiometric coefficients of the oxidation reaction
    by dichromate, given a dictionary of a molecule's constituent atoms and their 
    counts, as well as the number of negative charge, if any.
    
    This function is based on the oxidation of organic materials by dichromate in 
    an acid solution, as in a typical COD test of water or wastewater samples.
    Only C, H, O, N, S, P are considered active in the reaction.

    Parameters
    ----------
    atoms : dict[str, int or float]
        Dictionary of atoms and their counts, [-].
    charge : int or float
        Charge of the molecule.
    MW : float, optional
        Molecular weight of chemical, used only if `missing_handling` is 'Ash',
        [g/mol]
    missing_handling : str, optional
        How to handle compounds which do not appear in the stoichiometric
        reaction below. If 'elemental', return those atoms in the monatomic
        state; if 'ash', converts all missing attoms to 'Ash' in the output at
        a `MW` of 1 g/mol, [-]


    Returns
    -------
    stoichiometry : dict[str, float]
        Stoichiometric coefficients of the redox reaction. May inlcude the following
        keys for complete oxidation: 'H2O', 'CO2', 'NH4+', 'SO4-2', 'PO4-3'; 
        if `missing_handling` is 'elemental' can include the other elements; 
        if `missing_handling` is 'ash', Ash will be present in the output 
        if the compounds whose reactions are not included here. 
        'Cr2O7-2' is always present, with negative values indicating dichromate is
        required/consumed. [-]

    .. note::
        The stoichiometry is given by:
            .. math::
                C_n H_a O_b N_c S_d P_e^{f-} + xCr_2O_7^{2-} + (8x+c-2d-3e+f)H^{+} 
                    -> nCO_2 + 2xCr^{3+} + cNH_4^{+} + dSO_4^{2-} + ePO_4^{3-} + (b+7x-2n-4d-4e)H_2O
            .. math::
                x = \frac{4n+a-2b-3c+6d+5e+f}{6}
    
        Also included in the results is the moles of Cr2O7-2 required per mole of
        the mixture of the molecule.

        All products are in aqueous solution.
        
        Atoms not in ['C', 'H', 'N', 'O', 'S', 'P'] are returned as pure species; 
        i.e. sodium hydroxide produces water and pure Na.
        
    Examples
    --------
    >>> # Acetate in COD test:
    >>> cod_test_stoichiometry({'C': 2, 'H':3, 'O':2}, -1)
    {'Cr2O7-2': -1.3333333333333333,
     'H+': -11.666666666666666,
     'Cr+3': 2.6666666666666665,
     'CO2': 2,
     'H2O': 7.333333333333332}
    '''
    products = {}
    nC, nH, nO, nN, nS, nP = 0., 0., 0., 0., 0., 0.
    ne = - charge
    
    if 'C' in atoms:
        nC = atoms['C']
    if 'H' in atoms:
        nH = atoms['H']
    if 'O' in atoms:
        nO = atoms['O']
    if 'N' in atoms:
        nN = atoms['N']
    if 'S' in atoms:
        nS = atoms['S']
    if 'P' in atoms:
        nP = atoms['P']
    
    if nC <= 0 or nH <= 0:
        if not (len(atoms) == 1 and nH == 2):
            return {'Cr2O7-2': 0.}
    
    nCO2 = nC
    nNH4 = nN
    nSO4 = nS
    nPO4 = nP
    nCr2O7 = -(4*nC + nH - 2*nO - 3*nN + 6*nS + 5*nP + ne)/6
    nCr = -2*nCr2O7
    nH2O = nO - 7*nCr2O7 - 2*nC - 4*nS - 4*nP
    n_proton = 8*nCr2O7 - nN + 2*nS + 3*nP - ne

    if nCr2O7 != 0.0:
        products['Cr2O7-2'] = nCr2O7
    if n_proton != 0.0:
        products['H+'] = n_proton
    if nCr != 0.0:
        products['Cr+3'] = nCr
    if nCO2 !=  0.0:
        products['CO2'] = nCO2
    if nSO4 != 0.0:
        products['SO4-2'] = nSO4
    if nNH4 != 0.0:
        products['NH4+'] = nNH4
    if nPO4 != 0.0:
        products['PO4-3'] = nPO4
    if nH2O != 0.0:
        products['H2O'] = nH2O
        
    missing_handling = missing_handling.lower()
    if missing_handling == 'elemental':
        for atom, value in atoms.items():
            if atom not in dichromate_oxidizing_elements_set:
                products[atom] = value
    elif missing_handling == 'ash':
        cod_atoms = {i: atoms.get(i, 0) for i in dichromate_oxidizing_elements}
        MW = MW or molecular_weight(atoms)
        Ash = MW - molecular_weight(cod_atoms)
        if Ash/MW > 0.0001:
            products['Ash'] = Ash
    else:
        raise ValueError("Allowed values for `missing_handling` are 'elemental' and 'ash'.")
    return products

def electron_acceptor_cod(atoms, charge=0):
    if atoms == {'O':2}:
        return -1
    elif atoms == {'N':2}:
        return -1.5
    elif atoms == {'N':1, 'O':2} and charge == -1:
        return -1.5
    elif atoms == {'N':1, 'O':3} and charge == -1:
        return -2